#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'johnnycakes79'
"""
test_epys
----------------------------------

Tests for `epys` module.
"""

# import pytest
import os

from epys.read import *


this_dir, this_filename = os.path.split(__file__)
parent_dir = os.path.abspath(os.path.join(this_dir, os.pardir))
_dataRateFile = os.path.join(parent_dir, "test/data/data_rate_avg.out")
_powerFile = os.path.join(parent_dir, "test/data/power_avg.out")


# --- epys.read for data_rate_avg.out ---
def test_epys_read_datarate():
    data, header = read(_dataRateFile, meta=True)
    assert data.shape == (366, 52)
    assert len(header) == 20
    assert header['experiments'] == ['ANTENNA',
                                     'SSMM',
                                     'BELA',
                                     'ISA',
                                     'MERMAG',
                                     'MERTIS',
                                     'MGNS',
                                     'MORE',
                                     'PHEBUS',
                                     'SERENA',
                                     'SIMBIOSYS-STC',
                                     'SIMBIOSYS-HRIC',
                                     'SIMBIOSYS-VIHI',
                                     'MIXS-SIXS']


# --- epys.read for power_avg.out ---
def test_epys_read_power():
    data, header = read(_powerFile, meta=True)
    assert data.shape == (354, 15)
    assert len(header) == 20
    assert header['experiments'] == ['ANTENNA',
                                     'SSMM',
                                     'BELA',
                                     'ISA',
                                     'MERMAG',
                                     'MERTIS',
                                     'MGNS',
                                     'MORE',
                                     'PHEBUS',
                                     'SERENA',
                                     'SIMBIOSYS-STC',
                                     'SIMBIOSYS-HRIC',
                                     'SIMBIOSYS-VIHI',
                                     'MIXS-SIXS']



"""
This module processes mission analysis orbit files.
"""


import os
import logging
import numpy as np
from scipy import linalg
from spice import et2utc
from spice import furnsh


def orbitkernel(infile, outfile, cbody):
    """
    Make an SPK from the Mission Analysis Orbit file.
    """
    pass


def transform(fname, metadata=False):
    """
    This function reads a Mission Analysis Orbit file and performs a matrix
    transformation on it. Currently only from the Mercury Equatorial frame to
    the Earth Equatorial frame.

    :param fname: The path to the orbit file.
    :type fname: str.
    :param metadata: Flag to return the metadata dictionary
    :type state: bool.
    :returns:  numpy.array -- the return code.
    """

    furnsh("/Users/jmcaulif/Code/naif/generic/lsk/naif0010.tls")

    logging.basicConfig(level=logging.INFO)

    mdata = {}
    data = {}

    with open(fname, 'r') as fh:
        for line in fh:
            t, x, y, z, vx, vy, vz = [float(x) for x in line.split()]

            T = np.array([[0.98159386604468, 0.19098031873327, 0.0],
                          [-0.16775718426422, 0.86223242348167,
                           0.47792549108063],
                          [0.09127436261733, -0.46912873047114,
                           0.87840037851502]])

            Tinv = linalg.inv(T)

            r = np.array([[x, y, z]])
            v = np.array([[vx, vy, vz]])

            r_new = Tinv.dot(r.T).T
            v_new = Tinv.dot(v.T).T

            x, y, z = r_new[0]
            vx, vy, vz = v_new[0]

            t = et2utc(t * 86400, 'isoc', 2)

            print("{} {:9.2f} {:9.2f} {:9.2f} {:9.6f} {:9.6f} {:9.6f}".
                  format(t, x, y, z, vx, vy, vz))

    fh.close()

    if metadata:
        return data, mdata
    else:
        return data


def demotransform():
    """
    This function can be used to quickly get some data back for testing.
    It uses a pre-defined test orbit file.
    """

    # Grab the working directory and current file name by splitting
    # the os.path value.
    this_dir, this_filename = os.path.split(__file__)

    # Get the path to the parent directory for the current working directory.
    parent_dir = os.path.abspath(os.path.join(this_dir, os.pardir))

    # Build the path to the sample data files.
    samplefile = os.path.join(parent_dir,
                              "sample_data/8848MPO_DF50011_inMME2000.d40")

    # Run the test file through epys.read and save returned object to 'data'.
    # Ask for the return of the 'metadata' and save to 'meta'.
    data = transform(samplefile, metadata=False)

    # Return 'data' and 'meta' to the caller.
    return data

if __name__ == '__main__':
    demotransform()



class edf:

    content = []
    size = 0

    def __init__(self, fname):
        print("reading edf...")
        with open(fname, 'r') as fh:
            for line in fh:
                self.content.append(line)
        self.content = self.content[1:]
        self.size = len(self.content)

    def __str__(self):
        return self.content[0]


class erf:

    def __init__(self, fname):
        print("reading erf...")
        with open(fname, 'r') as fh:
            for line in fh:
                print line


class evf:

    def __init__(self, fname):
        print("reading evf...")
        with open(fname, 'r') as fh:
            for line in fh:
                print line


class itl:

    def __init__(self, fname):
        print("reading itl...")
        with open(fname, 'r') as fh:
            for line in fh:
                print line



# -*- coding: utf-8 -*
"""
This module contains a number of functions to read and parse various EPS/MAPPS
input and output data file formats and return the information (usually) in the
form of a pandas dataframe.
"""

import copy
import re
import pandas as pd
import numpy as np
from epys.utils import plotly_prep, background_colors
from datetime import datetime, timedelta
import logging
from plotly.graph_objs import Data, Layout, Figure, XAxis, YAxis
import plotly.plotly as py


class epstable:
    """
    This is the base class for data tables in epys. Upon instantiation it is
    provided with a filename which is read into a header dictionary and a
    Pandas dataframe. The class has a number of methods for querying its
    characteristics, plotting its contents and merging other tables into it.
    """
    def __init__(self, fname, columns=False):
        """
        This constructor method initialises the epstable object.

        :param fname: input filename
        :type fname: str
        :param columns: user supplied list of column headings
        :type columns: list or tuple
        :returns: epstable object consisting of a header dictionary and Pandas dataframe
        """
        self.data, self.header = read(fname, meta=True, columns=columns)

    # def __str__(self):
    #     pass

    # def __repr__(self):
    #     return self.header

    def range(self):
        """
        This method returns the time range of the dataframe as pandas Timestamp objects
        """
        return self.data.index[0], self.data.index[-1]

    def range_str(self):
        """
        This methond returns the time range of the dataframe as strings
        """
        return str(self.data.index[0]), str(self.data.index[-1])

    def range_jd(self):
        """
        This method returns the time range of the dataframe as julian date floats
        """
        return self.data.index[0].to_julian_date(), self.data.index[-1].to_julian_date()

    def range_datetime(self):
        """
        This method returns the time range of the dataframe as datetime objects pandas.tslib.Timestamp
        """
        return self.data.index[0].to_datetime(), self.data.index[-1].to_datetime()

    def head(self, n=False):
        """
        This method returns the first n (default 5) rows of the dataframe
        """
        if n:
            return self.data.head(n=n)
        else:
            return self.data.head()

    def tail(self, n=False):
        """
        This method returns the last n (default 5) rows of the dataframe
        """
        if n:
            return self.data.tail(n=n)
        else:
            return self.data.tail()

    def thin(self):
        """
        This method removes columns with only zeros.
        """
        self.data = self.data.ix[:, (self.data != 0).any(axis=0)]

    def iplot(self, selection=False, limits=False, title=False,
              x_title=False, x_range=False,
              y_title=False, y_range=False,
              showlegend=True, bg_alpha=False):
        """
        This method pre-processes the table object for plotting.

        :param selection: a sub-selection of columns to plot
        :type selection: list
        :param limits: start and end times of data to plot
        :type limits: datetime
        :param title: plot title
        :type title: str
        :param x_title: x-axis title
        :type x_title: str
        :param x_range: x-axis plot range
        :type x_range: list
        :param y_title: y-axis title
        :type y_title: str
        :param y_range: y-axis plot range
        :type y_range: list
        :param showlegend: flag to show the plot legend or not
        :type showlegend: boolean
        :param bg_alpha: factor for transparency of seasonal background
        :type bg_alpha: float (default=0.3)
        :returns:
        """

        if selection:
            if selection.__class__.__name__ == 'str':
                selection = re.split(' |,', selection)
                selection = [x for x in selection if x != '']
            cols = []
            for term in selection:
                results = [x for x in self.data.columns if term in x]
                for result in results:
                    cols.append(result)
            data_to_plot = self.data[cols]
        else:
            data_to_plot = self.data

        plot(data_to_plot, limits=limits, title=title,
            x_title=x_title, x_range=x_range,
            y_title=y_title, y_range=y_range,
            showlegend=showlegend, bg_alpha=bg_alpha)

    def join(self, df_to_join, in_place=False):
        """
        This method joins a second table object to this one

        :param df_to_join: table object to joint to this one
        :type df_to_join: epstable
        :param in_place: flag to join in place or return new instance
        :type in_place: boolean
        :returns:
        """
        # try:
        if in_place:
            self.data = merge_dataframes(self.data, df_to_join.data)
            try:
                self.data = self.data.sortlevel(axis=1)
            except:
                print('Warning: can only sort by level with a hierarchical index.')
        else:
            table_copy = copy.deepcopy(self)
            table_copy.data = merge_dataframes(table_copy.data, df_to_join.data)
            return table_copy
        # except:
        #     print('Ooops: that didn\'t work! Make sure your column ' +
        #           'and row indices match if using MutltiIndex. ' +
        #           'Or maybe the columns you\'re adding already exist.')
        #     return

    def sub(self, table_to_subtract, in_place=False):
        """
        This method subtracts the contents of second table from the
        contents of this one. Be careful with this... it's not been widely
        tested, I'm not sure what happens if the data tables have different
        column numbers and/or names.

        :param table_to_subtract: data table to subtract from this one
        :type table_to_subtract: epstable
        :param in_place: flag to perform the subtraction on this instance or return a new instance
        :type in_place: boolean
        :returns:
        """

        if in_place:
            self.data = self.data.sub(table_to_subtract.data)
        else:
            table_copy = copy.deepcopy(self)
            table_copy.data = table_copy.data.sub(table_to_subtract.data)
            return table_copy


class powertable(epstable):

    def __init__(self, fname):
        self.data, self.header = read(fname, meta=True)
        self.columns = zip(self.header['headings'], self.header['units'])
        self.instruments = self.header['experiments']

    def select(self, args, chop=False):
        """
        This function __does_something_unbelievable__

        :param level1: _add_description_here_
        :type level1: _add_type_here_
        :param level2: _add_description_here_
        :type level2: _add_type_here_
        :param level3: _add_description_here_
        :type level3: _add_type_here_
        :param chop: _add_description_here_
        :type chop: _add_type_here_
        :returns:
        """

        if chop:
            self.data = self.data[args]
        else:
            table_copy = copy.deepcopy(self)
            table_copy.data = table_copy.data[args]
            return table_copy


class datatable(epstable):

    def __init__(self, fname):
        """
        This function __does_something_unbelievable__

        :param fname: _add_description_here_
        :type fname: _add_type_here_
        :returns:
        """
        # read in the data
        self.data, self.header = read(fname, meta=True)
        # define experiments list
        self.instruments = self.header['experiments']
        # correct poorly- or un-named columns...
        unit_swaps = [('Kbits', 'kbit'), ('/sec', '/s'), ('Gbits', 'Gbit')]
        for swap in unit_swaps:
            for u in range(len(self.header['units'])):
                if swap[0] in self.header['units'][u]:
                    self.header['units'][u] = self.header['units'][u].replace(swap[0], swap[1])
        self.columns = zip(self.header['headings'], self.header['units'])
        cols = self.columns[1:]
        arrays = [[x[0].split()[0] for x in cols],
                  [x[0].split()[1] for x in cols],
                  [x[1] for x in cols]]
        self.data.columns = pd.MultiIndex.from_arrays(arrays)
        self.data = self.data.sortlevel(axis=1)
        # from each 'Accum' column produce a 'Volume column'
        for inst in self.instruments:
            try:
                self.data[inst, 'Volume', 'Gbit'] = self.data[inst, 'Accum', 'Gbit'].sub(self.data[inst, 'Accum', 'Gbit'].shift(), fill_value=0)
            except:
                print('The conversion of \'Accum\' to \'Volume\' didn\'t work for \'{}\'.'.format(inst))
        self.data = self.data.sort_index(axis=1)

    def select(self, level1, level2, level3, chop=False):
        """
        This function __does_something_unbelievable__

        :param level1: _add_description_here_
        :type level1: _add_type_here_
        :param level2: _add_description_here_
        :type level2: _add_type_here_
        :param level3: _add_description_here_
        :type level3: _add_type_here_
        :param chop: _add_description_here_
        :type chop: _add_type_here_
        :returns:
        """
        if not level1:
            level1 = slice(None)
        if not level2:
            level2 = slice(None)
        if not level3:
            level3 = slice(None)
        if chop:
            self.data = self.data.loc[slice(None), (level1, level2, level3)]
        else:
            table_copy = copy.deepcopy(self)
            table_copy.data = table_copy.data.loc[slice(None), (level1, level2, level3)]
            return table_copy


def plot(data, limits=False, title=False, x_title=False, x_range=False,
        y_title=False, y_range=False, showlegend=True, bg_alpha=False,
        yaxis1_limit=7000):
    """
    This function __does_something_unbelievable__

    :param limits: _add_description_here_
    :type limits: _add_type_here_
    :param title: _add_description_here_
    :type title: _add_type_here_
    :param x_title: _add_description_here_
    :type x_title: _add_type_here_
    :param x_range: _add_description_here_
    :type x_range: _add_type_here_
    :param y_title: _add_description_here_
    :type y_title: _add_type_here_
    :param y_range: _add_description_here_
    :type y_range: _add_type_here_
    :returns:
    """

    multiaxis = False

    if not limits:
        limits = [data.index[0], data.index[-1]]

    try:
        start = data.index.searchsorted(limits[0])
        end = data.index.searchsorted(limits[1])
    except:
        print('TypeError: \'limits\' should be a list of datetime objects, for now... sorry!')
        return

    if start == end:
        print('OOOPS: Your requested range appears to be outside the range of the data set:')
        print('data set > {} to {}'.format(str(data.index[0]), str(data.index[-1])))
        print('limits     > {} to {}'.format(str(limits[0]), str(limits[1])))
        return
    else:
        # Process the <i>power_data</i> dataframe into a <b>plotly</b>-ready <i>dictionary-list</i>.
        if data[start:end].shape[0] * data[start:end].shape[1] > 400000:
            print('Plotly will have issue with plotting so many points.')
            print('Suggestions: set plot limits or downsample data')
            return
        else:
            data_plotly, y_title_backup = plotly_prep(data[start:end])

            if y_title_backup.__class__.__name__ == 'tuple':
                multiaxis = True
                yaxis1_limit = max([max(x['y']) for x in data_plotly if x['yaxis'] == 'y1'])
                yaxis1_limit += yaxis1_limit * 0.1
            else:
                yaxis1_limit = max(data.max().tolist())
                yaxis1_limit += yaxis1_limit * 0.1

            # Ugly build for background color data...
            bg_colors = background_colors(top_limit=yaxis1_limit, bg_alpha=bg_alpha)

            # Create a plotly Data Graph Object from the plotly-ready
            # <b>power_data_plotly</b> dictionary.
            plot_data = Data(bg_colors + data_plotly)

            # Prepare plotly Graph Objects for the plot axes, layout and
            # figure  using data from <b>header</b>.
            if not title:
                title = ''

            if not x_title:
                x_title = ''
            if not x_range:
                x_range = [limits[0], limits[1]]
            xaxis = XAxis(title=x_title, range=x_range)

            if multiaxis:
                yaxis = YAxis(title=y_title_backup[0], range=[0, yaxis1_limit])
                yaxis2 = YAxis(title=y_title_backup[1],
                             overlaying='y',
                             side='right')
                # legend = Legend(x=1, y=1)
                layout = Layout(title=title, yaxis1=yaxis, yaxis2=yaxis2, xaxis1=xaxis,
                                showlegend=showlegend)
            else:
                if not y_title:
                    y_title = y_title_backup
                if not y_range:
                    y_max = max(data.max().tolist())
                    y_range = [0, y_max + y_max * 0.1]
                    # y_range = [0, 600]
                yaxis = YAxis(title=y_title, range=y_range)
                layout = Layout(title=title, yaxis1=yaxis, xaxis1=xaxis,
                                showlegend=showlegend)

            fig = Figure(data=plot_data, layout=layout)

            #Plot with <b>plotly</b>.
            py.iplot(fig, layout=layout)


def merge_dataframes(into_df, from_df):
    """
    This function merges two pandas data frames. The inital purpose was
    to merge sparse power and data downlink budgets into non-sparse EPS/MAPPS
    data_rate_avg.out and power_avg.out dataframes.

    :param into_df: power or data rate dataframe.
    :type into_df: pandas dataframe
    :param from_df: power or data downlink budget.
    :type from_df: pandas dataframe
    :returns: a merged dataframe with redundant NaN rows removed.
    """

    into_df_cols = into_df.columns.tolist()
    from_df_cols = from_df.columns.tolist()
    merged = pd.merge(from_df, into_df, how='outer', left_index=True, right_index=True)
    merged = merged[from_df_cols + into_df_cols]
    from_df_new = merged[from_df_cols]
    current_values = from_df.values[0]

    for i in xrange(merged.shape[0]):
        if np.isnan(from_df_new.values[i]).all():
            from_df_new.values[i] = current_values
        else:
            current_values = from_df_new.values[i]

    merged[from_df_cols] = from_df_new
    merged = merged.truncate(before=into_df.index.values[0],
                             after=into_df.index.values[-1])

    # build index of NaN rows for removal in the return statement
    inds = pd.isnull(merged[into_df_cols]).all(1).nonzero()[0]

    return merged.drop(merged.index[inds])


def parse_header(fname):
    """
    This function takes as input an EPS/MAPPS input or output data file and
    attempts to parse the header and return a key-value dictionary of meta
    data.

    :param fname: EPS/MAPPS input or output data file name
    :type fname: str
    :returns: key-value dictionary of meta data.
    """

    logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger(__name__)

    header = {}
    # data = []
    post_process = False
    experiments = []

    with open(fname, 'r') as fh:
        for line in fh:
            if 'len' in header:
                header['len'] += 1
            else:
                header['len'] = 0

            # Catch the header data and store it in dictionary.
            if re.match(r'#(.*):(.*)', line, re.M | re.I):
                keypair = line.strip('#\n').split(':')
                header[keypair[0].strip()] = keypair[1].strip()
                if 'Output Filename' in header:
                    if header['Output Filename'].split('_')[0] == 'data':
                        file_type = 'data'
                    elif header['Output Filename'].split('_')[0] == 'power':
                        file_type = 'power'
                    else:
                        print('ERROR: The input file is not of recognised type.')
                continue

            # Catch the reference date and add to dictionary.
            if re.match(r'Ref_date:(.*)', line, re.M | re.I):
                keypair = line.strip('#\n').split(':')
                ref_date = datetime.strptime(keypair[1].strip(), "%d-%b-%Y")
                header['Reference Date'] = ref_date
                continue

            # Catch the experiment names to list - data rate file_type
            if re.match(r'(.*)\<(.*)\>', line, re.M | re.I):
                for i in line.replace('.', '').replace('>', '').split():
                    experiments.append(i.replace('<', '').replace('_', '-'))
                header['experiments'] = experiments
                continue

            # Catch the column headers
            if re.match(r'Elapsed time(.*)', line, re.M | re.I):
                _headings = line.split()
                _headings[0:2] = [' '.join(_headings[0:2])]
                _headings = [h.replace('_', '-') for h in _headings]
                # if input is data rate file add experiments to headers
                if file_type == 'data':
                    for j in range(1, 3):
                        _headings[j] = experiments[0] + ' ' + _headings[j]
                    for j in range(3, 5):
                        _headings[j] = experiments[1] + ' ' + _headings[j]
                    x = 2
                    for j in range(5, len(_headings), 4):
                        for h in range(4):
                            _headings[j + h] = experiments[x] + \
                                ' ' + _headings[j + h]
                        x = x + 1
                if file_type == 'power':
                    experiments = _headings[2:]
                    header['experiments'] = experiments
                header['headings'] = _headings
                continue

            # Catch the units line and process ...
            if re.match(r'ddd_hh:mm:ss(.*)', line, re.M | re.I):
                _units = line.replace('(', '').replace(')', '').split()
                units = _units[0:2]
                for u in range(2, len(_units)):
                    if 'sec' in _units[u]:
                        units.append(_units[u])
                        units.append(_units[u])
                    else:
                        units.append(_units[u])
                post_process = True
                header['units'] = units
                continue

            if post_process:
                # Raise an error if the the length of 'units' is not equal
                # to the length of '_headings'.
                if len(_headings) != len(units):
                    logger.ERROR("ERROR: The number of headings does not ",
                                 "match the number of units!")

            # if start of data close file and return header ...
            if re.match(r'[0-9]{3}_[0-9]{2}:[0-9]{2}:[0-9]{2}(.*)',
                        line, re.M | re.I):
                fh.close()
                return header
            # ... same but for data rate / power budget file
            if re.match(r'[0-9]{2}-[0-9]{3}T[0-9]{2}:[0-9]{2}:[0-9]{2}(.*)Z(.*)',
                        line, re.M | re.I):
                fh.close()
                return header


def parse_time(*arg):
    """
    This function is a catch all for different time parsing methods.

    :param *arg: date string
    :returns: a datetime object
    """
    if len(arg) == 1:  # 'probably' coming from power or data budgets 24-084T05:00:00.000Z
        year_doy_time = arg[0]
        if re.match(r'[0-9]{2}-[0-9]{3}T[0-9]{2}:[0-9]{2}:(.*)Z(.*)',
                    year_doy_time, re.M | re.I):
            ref_date = datetime(int(year_doy_time.split('-')[0]) + 2000, 1, 1)
            days = int(year_doy_time.split('-')[1].split('T')[0])
            if days > 60 and ((int(year_doy_time.split('-')[0]) + 2000) % 4) == 0:
                days -= 1
            hours = int(year_doy_time.split('-')[1].split('T')[1].split(':')[0])
            minutes = int(year_doy_time.split('-')[1].split('T')[1].split(':')[1])
            seconds = int(round(float(year_doy_time.split('-')[1].split('T')[1].split(':')[2][:-1])))
            dtdelta = timedelta(days=days, hours=hours, minutes=minutes, seconds=seconds)
            return ref_date + dtdelta
        else:
            print('Error: Can\'t recognise time string format of {}.'.format(year_doy_time))
            return 1
    elif len(arg) == 2:  # 'probably' coming from power or data rate outfiles.
        days_time, ref_date = arg
        if re.match(r'[0-9]{3}_[0-9]{2}:[0-9]{2}:[0-9]{2}(.*)',
                    days_time, re.M | re.I):
            days, time = days_time.split('_')
            hours, minutes, seconds = time.split(':')
            time = ref_date + timedelta(days=int(days), hours=int(hours),
                                        minutes=int(minutes), seconds=float(seconds))
            return time
        else:
            print('Error: Can\'t recognise time string format of {}.'.format(days_time))
            return 1
    else:
        print('Error: Can\'t handle {} arguments.'.format(len(arg)))
        return 1


def read(fname, meta=False, columns=False):
    """
    This function reads any one of a number of EPS input/output files and
    returns the data in a pandas dataframe. The file metadata can also be
    returned if requested.

    :param fname: The path to the power_avg.out or data_rate_avg.out
    :type fname: str.
    :param meta: Flag to return the header dictionary
    :type meta: bool.
    :returns: pandas dataframe -- the return code.
    """
    header = {}
    header = parse_header(fname)
    if 'Output Filename' in header:
        try:
            data = pd.read_table(fname, skiprows=header['len'], header=None,
                                 names=header['headings'], sep=r"\s*", engine='python')
            data['Elapsed time'] = [parse_time(x, header['Reference Date']) for x in data['Elapsed time']]
            data = data.set_index(['Elapsed time'])
            data.index.names = ['Date']
            data = remove_redundant_data(data)
            if meta:
                return data, header
            else:
                return data
        except:
            print('Error: Didn\'t recognise file format...')
            return 1
    else:
        try:
            budget = pd.read_table(fname, header=None, comment='#', sep=r"\s*",
                                   # names=columns,
                                   engine='python')
            budget = budget[budget.ix[:, 1].notnull()]
            budget.ix[:, 0] = [parse_time(x) for x in budget.ix[:, 0]]
            budget.rename(columns={0: 'date'}, inplace=True)
            budget = budget.set_index(['date'])
            if columns:
                if len(columns) != len(budget.columns):
                    print('Error: \'columns\' length not equal to number of columns')
                else:
                    budget.columns = columns
            if meta:
                return budget, header
            else:
                return budget
        except:
            print('Error: Didn\'t recognise file format...')
            return 1


def remove_redundant_data(df):
    """
    This function reduces the size of a dataframe by reducing blocks of
    sequential identical data lines greater than 2 to only the earliest
    and latest.

    :param df: pandas dataframe
    :type df: pandas dataframe
    :returns: a smaller pandas dataframe
    """
    deletes = [False]  # we wanna keep the first row ...
    for i in range(df.shape[0] - 2):
        deletes.append(df.irow(i).tolist() == df.irow(i + 1).tolist()
                    == df.irow(i + 2).tolist())
    deletes.append(False)  # ... and the last row.
    keeps = [not i for i in deletes]  # flip the list
    if keeps.count(False) != 0:
        print('{} redundant lines removed'.format(keeps.count(False)))
    return df[keeps]



"""
This module does things with maps and images.
"""
from __future__ import print_function
from wand.image import Image
from lxml import etree
from datetime import datetime
from events import getMonth
import sys
import os
import re
import zipfile
import shutil
import utils
import binascii

try:
    from pykml.factory import KML_ElementMaker as KML
    NOKMLLIB = False
except:
    NOKMLLIB = True


def _makeTime(time):
    """
        extracts time from an input lineand returns a datetime object.
    """
    date, time = time.split('_')
    day, month, year = date.split('-')
    month = getMonth(month)
    hour, minute, second = time.split(':')
    return datetime(int(year), int(month), int(day),
                    int(hour), int(minute), int(second))


def _swathCenter(line):
    """
        calculate center of swath
    """
    lats = [float(line[3]), float(line[5]), float(line[7]), float(line[9])]
    lngs = [float(line[4]), float(line[6]), float(line[8]), float(line[10])]

    clat = min(lats) + (max(lats) - min(lats)) / 2.
    clng = min(lngs) + (max(lngs) - min(lngs)) / 2.

    return '{}'.format(clat), '{}'.format(clng)


def _buildSwath(line, data, polyalt=5000):
    """
    So far, an ugly hack on building the KML elements
    """

    # split input line
    line = line.split()

    # add polygon altitude to end of line tuple
    line.append(polyalt)

    # parse the time element of the line
    dt = _makeTime(line[1])
    date = dt.date()
    time = dt.time()

    # define time format string
    format = "%Y-%m-%dT%H:%M:%SZ"

    # ensure longitude is between -/+ 180 degrees
    for i in [4, 6, 8, 10]:
        if float(line[i]) > 180.0:
            val = float(line[i]) - 360.0
            line[i] = str(val)

    # build the vertices of the swath (remember the first vertex has to
    # repeat at the end.
    vertices = []
    for c in [3, 5, 7, 9, 3, 5]:
        vertices.append(",".join([line[i] for i in [c + 1, c, -1]]))

    # get center of swath
    clat, clng = _swathCenter(line)

    # create an image placemark for the kml
    image = KML.Placemark(
        # define name based on experiment and filter/channel
        KML.name('{}: {}'.format(data['experiment'], data['filter'])),
        # define description
        # TODO: come up with a more flexible way of doing this...
        KML.description(
            'Orbit no.:                          {}\n'.format(
                data['orbit']),
            'Pericenter time (UTC):              {}\n'.format(
                data['pericenter time'].replace('_', ' ')),
            'First image time (UTC):             {}\n'.format(
                data['first image time'].replace('_', ' ')),
            'First image time (from pericenter): {}\n'.format(
                data['first image time from pericenter'].replace('_',
                                                                 ' ')),
            'Last image time (UTC):              {}\n'.format(
                data['last image time'].replace('_', ' ')),
            'Last image time (from pericenter):  {}\n\n'.format(
                data['last image time from pericenter'].replace('_', ' ')),
            'Image sequence:                     {}\n'.format(line[0]),
            'Image date:                         {}\n'.format(date),
            'Image time:                         {}\n'.format(time),
            'Orbit no.:                          {}\n\n'.format(
                data['orbit']),
            'Pericentre relative time:           {}\n'.format(
                line[2].replace('_', ' ')),
            'Duration:                           {}\n\n'.format(line[20]),
            'S/C altitude:                       {}\n'.format(line[21]),
            'S/C latitude:                       {}\n'.format(line[22]),
            'S/C longitude:                      {}\n'.format(line[23]),
            'S/C target elevation:               {}\n'.format(line[24]),
            'S/C target azimuth:                 {}\n\n'.format(line[25]),
            'Reflection Angle:                   {}\n'.format(line[27]),
            'Sun target elevation:               {}\n'.format(line[28]),
            'Sun target azimuth:                 {}\n'.format(line[29]),
            'Target phase:                       {}\n'.format(line[30]),
            'Target elongation:                  {}\n'.format(line[31]),
            'Local Time:                         {}\n'.format(line[32]),
            'Image smear:                        {}\n'.format(line[33]),
            'Mercury True Anomaly:               {}\n'.format(line[35])
        ),
        # specify appearance time
        KML.TimeSpan(
            #KML.begin(str(tempTime))
            KML.begin(dt.strftime(format))
        ),
        # the style for this swath has been mapped in <swath>stylemap
        KML.styleUrl('#{}stylemap'.format(data['filter'])),
        #KML.styleUrl('#{}1'.format(data['filter'])),
        # define where the 'eye' looks when this swath is double clicked
        KML.LookAt(
            KML.longitude(clng),
            KML.latitude(clat),
            KML.altitude('5000'),
            KML.heading('0'),
            KML.tilt('30'),
            KML.roll('0'),
            KML.altitudeMode('relativeToGround'),
            KML.range('1500000')
        ),
        # defined the geometry object that will hold the swath polygon
        KML.MultiGeometry(
            # defined the swath polygon
            KML.Polygon(
                #KML.tessellate('1'),
                KML.altitudeMode('relativeToGround'),
                #KML.altitudeMode('clampedToGround'),
                KML.outerBoundaryIs(
                    KML.LinearRing(
                        KML.coordinates(
                            " ".join(vertices)
                        )
                    )
                )
            )
        )
    )

    return image


def _docSkel(name):
    name = re.sub('[^a-zA-Z0-9\n\.]', ' ', name.split('.')[0]).title()
    doc = KML.kml(
        KML.Document(
            KML.name(name),
            KML.open('1'),
            KML.visibility('1'),
            # uncomment the following if you want to hide the children
            #            KML.Style(
            #                KML.ListStyle(
            #                    KML.listItemType('checkHideChildren')
            #                ),
            #                id='check-hide-children',
            #                ),
            #            KML.styleUrl('#check-hide-children'),
        )
    )
    return doc


def _makeStyle(name, color):
    """
        Build swath pairs and map...
    """

    # style for the normal state of a swath
    stylen = KML.Style(
        KML.IconStyle(
            KML.scale('0.4'),
            KML.Icon(
                KML.href('http://maps.google.com/mapfiles/kml/shapes/star.png')
            )
        ),
        KML.LineStyle(
            KML.color('ff{}'.format(color)),
            KML.width(2.0)
        ),
        KML.LabelStyle(
            KML.color('990000ff'),
            KML.width('2')
        ),
        KML.PolyStyle(
            KML.color('00{}'.format(color)),
            KML.fill('1'),
            KML.outline('1')
        ),
        id='{}n'.format(name)
    )

    # style for the 'mouse-over' state of a swath
    styleh = KML.Style(
        KML.IconStyle(
            KML.scale('0.8'),
            KML.Icon(
                KML.href('http://maps.google.com/mapfiles/kml/shapes/star.png')
            )
        ),
        KML.LineStyle(
            KML.color('ffff4158'),
            KML.width(1.5)
        ),
        KML.LabelStyle(
            KML.color('990000ff'),
            KML.width('2')
        ),
        KML.PolyStyle(
            KML.color('fff7fff'),
            KML.fill('1'),
            KML.outline('1')
        ),
        id='{}h'.format(name)
    )

    # mapping of above styles
    stylem = KML.StyleMap(
        KML.Pair(
            KML.key('normal'),
            KML.styleUrl('#{}n'.format(name))
        ),
        KML.Pair(
            KML.key('highlight'),
            KML.styleUrl('#{}h'.format(name))
        ),
        id='{}stylemap'.format(name)
    )

    # Expand to make the style simpler...

    # kurl = 'http://maps.google.com/mapfiles/kml/shapes/star.png'

    #    style1 = KML.Style(
    #        KML.IconStyle(
    #            KML.scale('0.4'),
    #            KML.Icon(
    #                KML.href(kurl)
    #            )
    #        ),
    #        KML.LabelStyle(
    #            KML.color('990000ff'),
    #            KML.width('2')
    #        ),
    #        KML.LineStyle(
    #            KML.color('ff0000ff'),
    #            KML.width(2.0)
    #        ),
    #        KML.PolyStyle(
    #            KML.color('997f7fff'),
    #            KML.fill('1'),
    #            KML.outline('1')
    #        ),
    #        id='{}1'.format(name),
    #    )

    return stylen, styleh, stylem


def _buildKML(input, styles, experiments):
    """
        Put all the pieces together...
    """
    # create a KML file skeleton
    doc = _docSkel(input)

    # add the styles
    for style in styles:
        for part in style:
            doc.Document.append(part)

    # add the experiments
    for experiment in experiments.keys():
        doc.Document.append(experiments[experiment])

    return doc


def _writeKML(input, doc):
    """
        create and write to a KML file with the same name as
        the input file.
    """
    kmlName = "{}.kml".format(input.rsplit(".", 1)[0])
    outfile = open(kmlName, 'w')
    print(etree.tostring(doc, pretty_print=True), file=outfile)
    print("\nXML Doc written to {}.\n".format(kmlName))
    outfile.close()


def _gmapsusage(filename):
    """
        Construct a usage string.
    """
    usage = "{} <mapps_image_dump_file> [path_to_mapps_config_file]".format(
        filename)
    # Print usage string.
    print("\n[Usage]: python {}\n".format(usage))


def gmaps(input, configFile):
    """
       Check and deal with command line agruments.
    """

    if NOKMLLIB:
        print("\nOoops! 'gmaps' needs KML_ElementMaker from pykml.factory")
        print("       Try: pip install pykml\n")
    else:
        # # Check input arguments
        # if len(sys.argv) < 2 or len(sys.argv) > 3:
        #     # ... show usage hint...
        #     _gmapsusage(sys.argv[0])
        #     # ... exit!
        #     raise SystemExit(1)

        # input = sys.argv[1]
        # if sys.argv[2]:
        #     configFile = sys.argv[2]
        # else:
        #     configFile = False

        # create containers
        experiments = {}
        filters = {}
        currents = {}
        styles = []

        # Open input file for reading.
        infile = open(input, 'r')

        # Scan through the file line-by-line.
        # TODO: look into moving this for loop into a function
        for line in infile.readlines():
            if line.startswith('Trail'):
                break

            # TODO: Replace the crude pattern matching below with RegEx...
            if line.startswith('Experiment:'):
                expr = line.split(': ')[1].strip()
                if expr not in experiments:
                    experiments[expr] = KML.Folder(KML.name(expr.replace('_',
                                                                         ' ')),
                                                   KML.open('1'),
                                                   id='expr_{}'.format(expr))
                currents['experiment'] = expr

            if line.startswith('Swath:'):
                fltr = line.split(': ')[1].strip()
                if fltr not in filters:
                    filters[fltr] = KML.Folder(KML.name(fltr.replace('_', ' ')),
                                               KML.open('0'),
                                               KML.visibility('1'),
                                               KML.Style(KML.ListStyle(KML.listItemType('checkHideChildren')), id='check-hide-children'),
                                               KML.styleUrl('#check-hide-children'),
                                               id='fltr_{}'.format(fltr))
                    experiments[currents['experiment']].append(filters[fltr])
                currents['filter'] = fltr

            if line.startswith('Orbit:'):
                orbit = line.split()[1].strip()
                currents['orbit'] = orbit

            if line.startswith('Pericenter time (UTC):'):
                peric_time = line.split(': ')[1].strip()
                currents['pericenter time'] = peric_time

            if line.startswith('First image time (UTC):'):
                first_image_t = line.split(': ')[1].strip()
                currents['first image time'] = first_image_t

            if line.startswith('First image time (from pericenter):'):
                first_image_t_frm_peric = line.split(': ')[1].strip()
                currents['first image time from pericenter'] = first_image_t_frm_peric

            if line.startswith('Last image time (UTC):'):
                last_image_t = line.split(': ')[1].strip()
                currents['last image time'] = last_image_t

            if line.startswith('Last image time (from pericenter):'):
                last_image_t_frm_peric = line.split(': ')[1].strip()
                currents['last image time from pericenter'] = last_image_t_frm_peric

            # build an 'image' placemark element
            if line.startswith(' '):
                image = _buildSwath(line, currents)
                filters[currents['filter']].append(image)

        infile.close()

        # the styles for the different swaths
        colors = {}

        # if the MAPPS ini has been provided get colours from it.
        if configFile:
            inifile = open(configFile, 'r')
            for line in inifile.readlines():
                if '\swathColorName=' in line:
                    cHTML = line.rsplit("=#", 1)[1].strip()
                    cKML = '{}{}{}'.format(cHTML[4:6], cHTML[2:4], cHTML[0:2])
                    #print(cHTML, cKML)
                    bits = line.split('\\')[4]
                    #print(bits)
                    colors[bits] = cKML
        else:
            for fltr in filter.keys():
                cKML = binascii.b2a_hex(os.urandom(4))
                colors[fltr] = cKML
                # colors = ['641400E6', '6414F03C', '647828F0',
                #           '647828F0', '64F0FF14', '6478FFF0']

        for fltr in filters.keys():
            styles.append(_makeStyle(fltr, colors[fltr]))

        # build the KML file
        doc = _buildKML(input, styles, experiments)

        # TODO: fix schema checking...
        # schema_gx = Schema("kml22gx.xsd")
        # print(schema_gx.assertValid(doc))

        # write xml structure to kml file
        _writeKML(input, doc)


def splitter(originalFile, no_levels=3, zip=False, inichunk=False,
             demo=False):

    if demo:
        print('\nMAPPS Map Spliter **DEMO**, v0.1, 2014.')
    else:
        print('\nMAPPS Map Spliter, v0.1, 2014.')

    sys.stdout.write("\n   Importing original image...")
    sys.stdout.flush()
    img = Image(filename=originalFile)
    sys.stdout.write(" complete.\n")
    sys.stdout.flush()

    imgwidth = img.width
    imgheight = img.height

    if imgwidth / imgheight != 2:
        print('\n   Ooops!!! The Image Width to Height ratio should be 2!!!')
        return

    else:

        stem = originalFile.split('.')[0]

        if not os.path.exists(stem):
            os.makedirs(stem)
        else:
            print('\n   Uh-oh! The directory {} already exists.'.format(stem))
            if utils.yesno('   Do you want to replace it?'):
                shutil.rmtree(stem)
                os.makedirs(stem)
            else:
                return

        levels = range(1, no_levels + 1)
        for level in levels:

            print('\n   Processing Level {}'.format(level))

            split = 2 ** level
            segs = range(split)
            div = 1. / split

        for h in segs:
            for w in segs:
                w1 = int(imgwidth * div * w)
                w2 = int(imgwidth * div * (w + 1))
                h1 = int(imgheight * div * h)
                h2 = int(imgheight * div * (h + 1))

                imgtmp = img[w1:w2, h1:h2]
                # print(w1, w2, h1, h2)
                imgtmp.transform(resize='1440x720!')
                imgtmp.format = 'jpeg'

                hlevel = '{0:03d}'.format(h + 1)
                wlevel = '{0:03d}'.format(w + 1)
                saveas = os.path.join(stem, '{}_{}_{}_{}.jpg'.format(
                    stem, level, hlevel, wlevel))

                print('      Writing: {}'.format(saveas))
                imgtmp.save(filename=saveas)

                if imgtmp.width != 1440:
                    print('ERROR: image width = {}\n'.format(imgtmp.width))
                if imgtmp.height != 720:
                    print('ERROR: image height = {}\n'.format(imgtmp.height))

        # process input image
        img.transform(resize='1440x720')
        img.format = 'jpeg'
        img.save(filename=os.path.join(stem, '{}_0_001_001.jpg'.format(
            stem)))

        # create ini file segment
        if inichunk:
            utils.inifix(stem, no_levels)

        # zip output
        if zip:
                print('\n   Zipping output to {}.zip'.format(stem))
                zipf = zipfile.ZipFile('{}.zip'.format(stem), 'w')
                utils.zipdir('{}/'.format(stem), zipf)
                zipf.close()
                shutil.rmtree(stem)

        print('\nFinished!\n')

        return


def tabular(input, configFile):
    """
       Check and deal with command line agruments.
    """

    # create containers
    # experiments = {}
    filters = {}
    currents = {}
    styles = []
    first = True

    # Open input file for reading.
    infile = open(input, 'r')

    header = ['seqnr','image time (UTC)','peri reltime','p1 lat','p1 long',
              'p2 lat','p2 long','p3 lat','p3 long','p4 lat','p4 long',
              'type','p5 lat','p5 long','p6 lat','p6 long','p7 lat','p7 long',
              'p8 lat','p8 long','duration','altitude','SC latitude',
              'SC longitude','SC target elev','SC target  azimuth',
              'distance','reflection','Sun tgt elev','Sun tgt azimuth',
              'tgt phase','tgt elongation','local time','img smear',
              'tg phase','perihelion','attdata']

    # Scan through the file line-by-line.
    # TODO: look into moving this for loop into a function
    for line in infile.readlines():
        if line.startswith('Trail'):
            break

        # TODO: Replace the crude pattern matching below with RegEx...
        if line.startswith('Experiment:'):
            expr = line.split(': ')[1].strip()
            currents['experiment'] = expr

        if line.startswith('Swath:'):
            fltr = line.split(': ')[1].strip()
            currents['filter'] = fltr

        if line.startswith('Orbit:'):
            orbit = line.split()[1].strip()
            currents['orbit'] = orbit

        if line.startswith('Pericenter time (UTC):'):
            peric_time = line.split(': ')[1].strip()
            currents['pericenter time'] = peric_time

        if line.startswith('First image time (UTC):'):
            first_image_t = line.split(': ')[1].strip()
            currents['first image time'] = first_image_t

        if line.startswith('First image time (from pericenter):'):
            first_image_t_frm_peric = line.split(': ')[1].strip()
            currents['first image time from pericenter'] = first_image_t_frm_peric

        if line.startswith('Last image time (UTC):'):
            last_image_t = line.split(': ')[1].strip()
            currents['last image time'] = last_image_t

        if line.startswith('Last image time (from pericenter):'):
            last_image_t_frm_peric = line.split(': ')[1].strip()
            currents['last image time from pericenter'] = last_image_t_frm_peric

        if line.startswith('seqnr') and first:
            print(','.join(currents.keys()+header))
            first = False

        # build an 'image' placemark element
        if line.startswith(' '):
            # image = _buildSwath(line, currents)
            # filters[currents['filter']].append(image)
            print(','.join(currents.values()+line.split()))

        infile.close()

        # the styles for the different swaths
        colors = {}

        # if the MAPPS ini has been provided get colours from it.
        if configFile:
            inifile = open(configFile, 'r')
            for line in inifile.readlines():
                if '\swathColorName=' in line:
                    cHTML = line.rsplit("=#", 1)[1].strip()
                    cKML = '{}{}{}'.format(cHTML[4:6], cHTML[2:4], cHTML[0:2])
                    #print(cHTML, cKML)
                    bits = line.split('\\')[4]
                    #print(bits)
                    colors[bits] = cKML
        else:
            for fltr in filter.keys():
                cKML = binascii.b2a_hex(os.urandom(4))
                colors[fltr] = cKML
                # colors = ['641400E6', '6414F03C', '647828F0',
                #           '647828F0', '64F0FF14', '6478FFF0']

        for fltr in filters.keys():
            styles.append(_makeStyle(fltr, colors[fltr]))

        # build the KML file
        # doc = _buildKML(input, styles, experiments)

        # TODO: fix schema checking...
        # schema_gx = Schema("kml22gx.xsd")
        # print(schema_gx.assertValid(doc))

        # write xml structure to kml file
        # _writeKML(input, doc)


# demo wrappers for testing the above...
def gmapsdemo():
    gmaps('imagedatadump.dat', '/Users/jmcaulif/Code/bepic/ESA/bepi_mapps_v65.ini')


def splitterdemo():
    shutil.rmtree("demo")
    splitter("demo.png", demo=True, no_levels=2)


def tabulardemo():
    tabular('../sample_data/imagedatadump.dat', '/Users/jmcaulif/Code/bepic/ESA/bepi_mapps_v65.ini')


if __name__ == '__main__':
    # gmapsdemo()
    # splitterdemo()
    tabulardemo()



#!/usr/bin/env python
"""
ePYs is a python library for the manipulation, processing and plotting
of the input and output files of ESA Experiment Planning Software (EPS).

.. WARNING::
   This is a very beta-project. It's not on PyPI and can't be installed
   via PIP.
"""

__author__ = 'Jonathan McAuliffe'
__email__ = 'watch.n.learn@gmail.com'
__version__ = '0.3.5'
__url__ = 'https://github.com/johnnycakes79/epys'

# from epys import classes, draw, events, maps, orbit, read, utils

# from epys.draw import planetsplot

from epys.read import epstable, datatable, powertable, read

from epys.utils import plotly_prep, background_colors



"""
This module provides a series of time/date utilities.
"""
from __future__ import print_function
from datetime import datetime, timedelta


# Classes
class line():
    """
        a class to hold information from a line from an
        EPS event file.
    """
    def __init__(self, string):
        self.vars = string.split('  ')
        self.time = extractTime(self.vars[0])
        self.tag = self.vars[1]
        self.count = tagnum(self.vars[2])

    def tagHas(self, *args):
        tagsplit = self.tag.split('_')
        for t in args:
            if t not in tagsplit:
                return False
        return True


class window():
    """
        a simple class that has a start time, stop time and duration
    """

    def __init__(self, start=0, stop=0):
        self.start = start
        self.stop = stop

    def duration(self):
        return self.stop - self.start


# class test_class():
#     """
#         a class to test functionality
#     """

#     def __init__(self,  **kwargs):
#         self.properties = kwargs

#     def set_kwp(self, k, v):
#         self.properties[k] = v

#     def get_kwv(self, key):
#         return self.properties.get(key, None)

#     def len(self):
#         return __len__(self)


class parameter():
    """
        an eps parameter class
    """

    def __init__(self, parameter, description, raw_type='', eng_type='',
                 default_value='', unit='', raw_limits=[], eng_limits=[],
                 parameter_values=[]):
        self.parameter = parameter
        self.description = description
        self.raw_type = raw_type
        self.eng_type = eng_type
        self.default_value = default_value
        self.unit = unit
        if len(raw_limits) in [0, 2]:
            self.raw_limits = raw_limits
        else:
            print("Limits can only have 2 values.")
        if len(eng_limits) in [0, 2]:
            self.eng_limits = eng_limits
        else:
            print("Limits can only have 2 values.")
        self.parameter_values = parameter_values

    # def len(self):
    #     return __len__(self)


# Time functions
def getMonth(month):
    """
        returns the number of the month if given a string
        and returns the name of the month if given and int
    """
    # months = ['January', 'February', 'March', 'April', 'May', 'June',
    #           'July', 'August', 'September', 'October' 'November',
    #           'December']
    mons = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug',
            'Sep', 'Oct', 'Nov', 'Dec']
    if type(month) == str:
        month = month[0:3]
        try:
            return mons.index(month) + 1
        except ValueError as err:
            print('Shall we list the months: {}'.format(err))
    if type(month) == int:
        try:
            return mons[month - 1]
        except IndexError as err:
            print('How many months are there in the year: {}'.format(err))


def extractTime(line):
    """
        extracts time from an event file input line
        and returns a datetime object.
    """
    dt = line.split(None, 1)[0]
    if len(dt) > 20:
        print(dt, 'Invalid dateTime string')
        return
    #print('extractTime : ', line )
    d, t = dt.split('_')
    day, month, year = d.split('-')
    month = getMonth(month)
    hour, minute, second = t.split(':')
    return datetime(int(year), int(month), int(day),
                    int(hour), int(minute), int(second))


def insertTime(line, time):
    """
        inserts a new time into an event file line.
    """
    dt = line[0:line.find(' ')]
    return line.replace(dt, time)


def constructTime(dt):
    """
        reconstructs an event file line time element
        from a datetime object.
    """
    if type(dt) != datetime:
        print('Invalid datetime object')
        return
    return (str(dt.day).zfill(2) + '-' + getMonth(dt.month) + '-'
            + str(dt.year) + '_' + str(dt.hour).zfill(2) + ':'
            + str(dt.minute).zfill(2) + ':' + str(dt.second).zfill(2))


def adjustTime(line, toChange, by):
    """
        adjusts the datetime element 'toChange' of an
        event line by 'by'.
    """
    adj = timedelta(**{toChange: by})
    dt = extractTime(line)
    new_dt = dt + adj
    return constructTime(new_dt)


# Line editor functions
def adjustLine(line, timeChange, timeBy, tagChange=None):
    """
        adjusts the datetime element of an event line and,
        if given, change the event tag element to 'tagChange'.
    """
    dt, tag, cntstr = line.split(None, 2)
    if tagChange is not None:
        tag = tagChange
    dt = adjustTime(dt, timeChange, timeBy)
    return '{}  {}  {}'.format(dt, tag, cntstr)


# Content test functions
def containsAll(line, *args):
    """
        tests to see if 'line' contains ALL of the
        tags in 'args'
    """
    #print('Line : {}Tags : {}'.format(line, args))
    for tag in args:
        #print('Search Tag : {}'.format(tag))
        # If any of the tags are not found 'line.find(tag)' will be -1.
        if (line.find(tag) < 0):
            #print('{} does not contain {}'.format(line.strip('\n'), tag))
            return (line.find(tag) > 0)
        #else:
            #print('{} contains {}'.format(line.strip('\n'), tag))
    return (line.find(tag) > -1)


def containsAny(line, *args):
    """
        tests to see if 'line' contains ANY of the
        tags in 'args'
    """
    for tag in args:
        if line.find(tag) > -1:
            return line.find(tag) > -1
    return line.find(tag) > -1


def containsNone(line, *args):
    """
        tests to see if 'line' contains 'NONE' of the
        tags in 'args'
    """
    return (not containsAny(line, *args))


# Event count function
def tagnum(line):
    """
        returns the current event count
    """
    return int(line[line.find('=') + 1:line.find(')')])


# Is line in the header?
def isHeader(line):
    """
        tests to see if 'line' is in the event file
        header
    """
    if containsAny(line, 'EVF Filename:', 'Generation Time:', 'Start_time:',
                   'End_time:', 'events in list)', '#', 'Include:',
                   'Init_value:'):
        return True
    elif len(line) < 3:
        return True
    else:
        return False


# Text wrapped output
def wrap(text, width):
    '''
    @summary: A word-wrap function that preserves existing line breaks
    and most spaces in the text. Expects that existing line
    breaks are posix newlines.
    @param text:
    @param width:
    @result:
    '''
    return reduce(lambda line, word, width=width: '%s%s%s' %
                  (line,
                   ' \n'[(len(line) - line.rfind('\n') - 1
                         + len(word.split('\n', 1)[0]
                               ) >= width)],
                   word
                   ),
                  text.split(' ')
                  )


# Main function
def main():
    """
        does nothing for now...
    """
    print('This is a random collection of functions... TBS - to be sorted.')


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*
"""
This module is very useful...
"""
from bisect import bisect_left
import os
import sys
# import spice
from plotly.graph_objs import Scatter


def all_same(items):
    return all(x == items[0] for x in items)


def background_colors(top_limit=7000, limits=False, bg_alpha=False):
    """
    This function builds a list of plotly Scatter traces for the power
    based seasons. Currently a manual hack.
    TODO: read in power file and build color traces.

    :param top_limit: maximum y-values of background traces.
    :type top_limit: int or float
    :returns: list of plotly Scatter objects
    """
    if not bg_alpha:
        bg_alpha = 0.2

    no_limit = u'''rgba(0, 255, 0, {})'''.format(bg_alpha)
    some_limit = u'''rgba(255,165,0, {})'''.format(bg_alpha)
    big_limit = u'''rgba(255,0,0, {})'''.format(bg_alpha)

    bg_colors = []
    bg_colors.append(Scatter(x=['2024-03-27 18:18:00', '2024-06-03 19:18:00'], y=[top_limit, top_limit], fill='tozeroy', mode='line', yaxis='y1', showlegend=False, fillcolor=no_limit))  # 60
    bg_colors.append(Scatter(x=['2024-06-03 19:18:00', '2024-06-09 20:18:00'], y=[top_limit, top_limit], fill='tozeroy', mode='line', yaxis='y1', showlegend=False, fillcolor=some_limit))  # 301
    bg_colors.append(Scatter(x=['2024-06-09 20:18:00', '2024-06-17 15:48:00'], y=[top_limit, top_limit], fill='tozeroy', mode='line', yaxis='y1', showlegend=False, fillcolor=big_limit))  # 336
    bg_colors.append(Scatter(x=['2024-06-17 15:48:00', '2024-06-23 17:18:00'], y=[top_limit, top_limit], fill='tozeroy', mode='line', yaxis='y1', showlegend=False, fillcolor=some_limit))  # 25
    bg_colors.append(Scatter(x=['2024-06-23 17:18:00', '2024-08-30 18:18:00'], y=[top_limit, top_limit], fill='tozeroy', mode='line', yaxis='y1', showlegend=False, fillcolor=no_limit))  # 60
    bg_colors.append(Scatter(x=['2024-08-30 18:18:00', '2024-09-05 19:48:00'], y=[top_limit, top_limit], fill='tozeroy', mode='line', yaxis='y1', showlegend=False, fillcolor=some_limit))  # 301
    bg_colors.append(Scatter(x=['2024-09-05 19:48:00', '2024-09-13 14:48:00'], y=[top_limit, top_limit], fill='tozeroy', mode='line', yaxis='y1', showlegend=False, fillcolor=big_limit))  # 336
    bg_colors.append(Scatter(x=['2024-09-13 14:48:00', '2024-09-19 16:48:00'], y=[top_limit, top_limit], fill='tozeroy', mode='line', yaxis='y1', showlegend=False, fillcolor=some_limit))  # 25
    bg_colors.append(Scatter(x=['2024-09-19 16:48:00', '2024-11-26 17:48:00'], y=[top_limit, top_limit], fill='tozeroy', mode='line', yaxis='y1', showlegend=False, fillcolor=no_limit))  # 60
    bg_colors.append(Scatter(x=['2024-11-26 17:48:00', '2024-12-02 18:48:00'], y=[top_limit, top_limit], fill='tozeroy', mode='line', yaxis='y1', showlegend=False, fillcolor=some_limit))  # 301
    bg_colors.append(Scatter(x=['2024-12-02 18:48:00', '2024-12-10 14:18:00'], y=[top_limit, top_limit], fill='tozeroy', mode='line', yaxis='y1', showlegend=False, fillcolor=big_limit))  # 336
    bg_colors.append(Scatter(x=['2024-12-10 14:18:00', '2024-12-16 15:48:00'], y=[top_limit, top_limit], fill='tozeroy', mode='line', yaxis='y1', showlegend=False, fillcolor=some_limit))  # 25
    bg_colors.append(Scatter(x=['2024-12-16 15:48:00', '2025-02-23 16:48:00'], y=[top_limit, top_limit], fill='tozeroy', mode='line', yaxis='y1', showlegend=False, fillcolor=no_limit))  # 60
    bg_colors.append(Scatter(x=['2025-02-23 16:48:00', '2025-03-01 18:18:00'], y=[top_limit, top_limit], fill='tozeroy', mode='line', yaxis='y1', showlegend=False, fillcolor=some_limit))  # 301
    bg_colors.append(Scatter(x=['2025-03-01 18:18:00', '2025-03-09 13:18:00'], y=[top_limit, top_limit], fill='tozeroy', mode='line', yaxis='y1', showlegend=False, fillcolor=big_limit))  # 336
    bg_colors.append(Scatter(x=['2025-03-09 13:18:00', '2025-03-15 15:18:00'], y=[top_limit, top_limit], fill='tozeroy', mode='line', yaxis='y1', showlegend=False, fillcolor=some_limit))  # 25
    bg_colors.append(Scatter(x=['2025-03-15 15:18:00', '2025-05-21 16:18:00'], y=[top_limit, top_limit], fill='tozeroy', mode='line', yaxis='y1', showlegend=False, fillcolor=no_limit))  # 60
    bg_colors.append(Scatter(x=['2025-05-21 16:18:00', '2025-05-27 17:18:00'], y=[top_limit, top_limit], fill='tozeroy', mode='line', yaxis='y1', showlegend=False, fillcolor=some_limit))  # 301
    bg_colors.append(Scatter(x=['2025-05-27 17:18:00', '2025-06-04 12:48:00'], y=[top_limit, top_limit], fill='tozeroy', mode='line', yaxis='y1', showlegend=False, fillcolor=big_limit))  # 336
    bg_colors.append(Scatter(x=['2025-06-04 12:48:00', '2025-06-08 00:00:00'], y=[top_limit, top_limit], fill='tozeroy', mode='line', yaxis='y1', showlegend=False, fillcolor=some_limit))  # 25

    return bg_colors


# def getclosest(myList, myNumber):
#     '''
#     Assumes myList is sorted. Returns closest value to myNumber.
#     If two numbers are equally close, return the smallest number.

#     :param myList:
#     :param myNumber:
#     :returns:
#     '''
#     pos = bisect_left(myList, myNumber)
#     if pos == 0:
#         return myList[0]
#     if pos == len(myList):
#         return myList[-1]
#     before = myList[pos - 1]
#     after = myList[pos]
#     if after - myNumber < myNumber - before:
#         return after
#     else:
#         return before


# def getorbelts(epoch, planet='MERCURY', spacecraft='MPO', verbose=False,
#                ltcorr='NONE', reframe='J2000'):
#     """
#     Main function that ...
#     """

#     spice.kclear()

#     # Load the kernels that this program requires.
#     spice.furnsh('epys.mk')

#     # convert starting epoch to ET
#     et = spice.str2et(epoch)

#     orbelts = []

#     if verbose:
#         print('ET Seconds Past J2000: {}'.format(et))
#         print('Date: {}'.format(spice.et2utc(et, 'C', 0)))

#     # Compute the apparent state of MPO as seen from Mercury in J2000
#     starg, ltime = spice.spkezr(spacecraft, et, reframe, ltcorr, planet)

#     # Let starg contain the initial state of a spacecraft relative to
#     # the center of a planet at epoch ET, and let GM be the gravitation
#     # parameter of the planet. The call

#     mu = planetmu(planet)
#     elts = spice.oscelt(starg, et, mu)

#     # elts.append(spice.et2utc(et,'ISOC', 0).split('T')[0])

#     orbelts.append(elts[0])
#     orbelts.append(elts[1] * spice.dpr())
#     orbelts.append(elts[2] * spice.dpr())
#     orbelts.append(elts[3] * spice.dpr())
#     if elts[4] * spice.dpr() > 180.:
#         orbelts.append(elts[4] * spice.dpr() - 360)
#     else:
#         orbelts.append(elts[4] * spice.dpr())
#     orbelts.append(elts[5] * spice.dpr())
#     orbelts.append(elts[0] - 2439)

#     # spice.kclear()

#     return orbelts


def yesno(question, default="yes"):
    '''
    Ask a yes/no question via raw_input() and return their answer.

    :param question: is a string that is presented to the user.
    :param default: is the presumed answer if the user just hits <Enter>. It must be "yes" (the default), "no" or None (meaning an answer is required of the user).
    :returns: The "answer" return value is one of "yes" or "no".
    '''
    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = raw_input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")


def zipdir(path, zip):
    """
    This function _add_summary_here_

    :param path: _add_description_here_
    :param zip: _add_description_here_
    :returns:
    """

    for root, dirs, files in os.walk(path):
        for file in files:
            zip.write(os.path.join(root, file))


def inifix(stem, no_levels):
    """
    This function _add_summary_here_

    :param stem: _add_description_here_
    :param no_levels: _add_description_here_
    :returns:
    """

    print('\n   Writing config chunk to {}.ini'.format(stem))
    iniChunk = open(os.path.join(stem, '{}.ini'.format(stem)), "w")
    iniChunk.write('***Update elements between asterisks***\n')
    iniChunk.write('[Background]\n')
    iniChunk.write('Definitions\\***00X***\\appearsInToolbar=true\n')
    iniChunk.write('Definitions\\***00X***\\autoNrOfLevels=true\n')
    iniChunk.write('Definitions\\***00X***\\backgroundType=ImageSet\n')
    iniChunk.write('Definitions\\***00X***\\baseImageName=',
                   '{}_%1_%2_%3.jpg\n'.format(stem))
    iniChunk.write('Definitions\\***00X***\\hasIcon=true\n')
    iniChunk.write('Definitions\\***00X***\\hasLegend=false\n')
    iniChunk.write('Definitions\\***00X***\\hasOffset=false\n')
    iniChunk.write('Definitions\\***00X***\\iconName=shdrelmap.xpm\n')
    iniChunk.write('Definitions\\***00X***\\imageHeight=720\n')
    iniChunk.write('Definitions\\***00X***\\imageWidth=1440\n')
    iniChunk.write('Definitions\\***00X***\\isActive=true\n')
    iniChunk.write('Definitions\\***00X***\\longitudeOffset=90\n')
    iniChunk.write('Definitions\\***00X***\\maxPolarLevels=2\n')
    iniChunk.write('Definitions\\***00X***\\menuAccelerator=',
                   '***set quick key combo***\n')
    iniChunk.write('Definitions\\***00X***\\menuTextLabel=',
                   '***set label name*** (***set quick key combo***)\n')
    iniChunk.write('Definitions\\***00X***\\nrOfLevels={}\n'.format(no_levels))
    iniChunk.write('Definitions\\***00X***\\toolbarMessageText=',
                   '{}\n'.format(stem.replace('_', ' ').upper()))
    iniChunk.write('Definitions\\***00X***\\menuStatusTipText=',
                   '{}\n'.format(stem.replace('_', ' ').upper()))
    iniChunk.close()


def planetmu(planet):

    # unit = μ (km3s−2)
    # source: http://en.wikipedia.org/wiki/Standard_gravitational_parameter

    planetmu = {'Sun': 132712440018, 'Mercury': 22032, 'Venus': 324859,
                'Earth': 398600.4418, 'Moon': 4902.8000, 'Mars': 42828,
                'Ceres': 63.1, 'Jupiter': 126686534, 'Saturn': 37931187,
                'Uranus': 5793939, 'Neptune': 6836529, 'Pluto': 871,
                'Eris': 1108}

    return planetmu[planet.title()]


def plotly_prep(df):
    """
    This fuction prepares a Pandas dataframe for plotting in Plotly:
        http://nbviewer.ipython.org/gist/nipunreddevil/7734529


    :param df: dataframe for pre-processing for plotly
    :type df: pandas dataframe
    :returns: plotly-ready list of dictionaries
    """
    if df.index.__class__.__name__ == "DatetimeIndex":
        # convert the index to MySQL Datetime like strings
        x = df.index.format()
        # zlternatively, directly use x, since DateTime index is np.datetime64
        # see http://nbviewer.ipython.org/gist/cparmer/7721116
        # x=df.index.values.astype('datetime64[s]')
    else:
        x = df.index.values

    lines = {}
    multidx = False
    keeps = []

    # if the columns header are tuples and not strings then we're
    # dealing with a multi-index
    if df.columns[0].__class__.__name__ == 'tuple':
        multidx = True

    if multidx:
        # exclude redundant information from the legend.
        for i in range(len(df.columns[0])):
            keeps.append(not all_same([key[i] for key in df]))

        if all_same(keeps) is True and keeps[0] is False:
            keeps = [not i for i in keeps]
            keeps[-1] = False

        # get the set of axes by making a list of the last element
        # of each key in the dataframe, then converting this to a
        # set and then back to a list.
        axes = list(set([key[-1] for key in df]))

    # build a list of dictionaries for each trace with meta data.
    for key in df:
        # if we're dealing with a multi-index assign this trace to the
        # appropriate axis by checking the last element of the key list
        # against the axes set.
        if multidx:
            axis = axes.index(key[-1]) + 1
        # define dict for this trace
        lines[key] = {}
        # assign the x values defined above to the dict key 'x'
        lines[key]["x"] = x
        # assign the values for this key to the dict key 'y'
        lines[key]["y"] = df[key].values
        if multidx:
            # if multi-index specify axis as defined above
            lines[key]["yaxis"] = 'y{}'.format(axis)
            # and build a name for the trace using the non-redundant parts
            # of the key. what this means is that if you are plotting 10 traces
            # all of the same variable then the name and unit of the variable
            # will be redundant and need not show up in the name which will
            # be used to populate the legend.
            lines[key]["name"] = ' '.join([k for k in key if keeps[key.index(k)]])
        else:
            # if we're not dealing with a multi-index the key will just be a string
            # and can bue used for the trace name.
            lines[key]["name"] = key

    if multidx:
        # if dealing with a multi-index flip the keeps list to define a
        # 'what-to-cut' list.
        cuts = [not i for i in keeps]
        # build a  y-axis title stem using the bits cut from the trace name.
        y_title = ' '.join([k for k in df.columns[0] if cuts[df.columns[0].index(k)]])
        if len(axes) == 2:
            # if we have 2 axes then build a tuple of the title stem and the units
            # wrapped in brackets.
            y_title = (y_title + ' [' + axes[0] + ']', y_title + ' [' + axes[1] + ']')
        if len(axes) > 2:
            # if we have more than 2 axes... panic. this isn't dealt with yet.
            print('Hmmm, more than 2 unit types...')
    else:
        # other wise leave it empty
        y_title = ''

    # build list of all traces ..
    lines_plotly = [lines[key] for key in df]
    # and return the list as the y-title.
    return lines_plotly, y_title



"""
This module makes pretty orbit graphics.
"""

from utils import getclosest, getorbelts, planetmu
from datetime import datetime
from PyAstronomy import pyasl
from PyAstronomy import constants as consts
import matplotlib.pyplot as plt
import numpy as np
import svgfig as svg
import inspect
import spice
import math
import os

from io import StringIO
import telnetlib
import socket

consts.setSystem('SI')


class Horizons(telnetlib.Telnet, object):

    MERCURY = 199
    VENUS = 299
    EARTH = 399
    MARS = 499
    JUPITER = 599
    SATURN = 699
    URANUS = 799
    NEPTUNE = 899
    PLUTO = 999

    def __init__(self, timeout=socket._GLOBAL_DEFAULT_TIMEOUT):
        super(Horizons, self).__init__("localhost", 6775, timeout)

    def open(self, host, port=0, timeout=socket._GLOBAL_DEFAULT_TIMEOUT):
        super(Horizons, self).open(host, port, timeout)
        # Disable pager on start
        self._check_main()
        self.sendline("page")

    def sendline(self, s=""):
        self.write(s.encode('ascii') + b"\n")

    def elements(self, body, start_date, end_date, delta):
        """Compute osculatory elements, selected a body.

        Columns available:
        0JDCT     Epoch Julian Date, Coordinate Time
        2EC     Eccentricity, e
        3QR     Periapsis distance, q (km)
        4IN     Inclination w.r.t xy-plane, i (degrees)
        5OM     Longitude of Ascending Node, OMEGA, (degrees)
        6W      Argument of Perifocus, w (degrees)
        7Tp     Time of periapsis (Julian day number)
        8N      Mean motion, n (degrees/sec)
        9MA     Mean anomaly, M (degrees)
        10TA     True anomaly, nu (degrees)
        11A      Semi-major axis, a (km)
        12AD     Apoapsis distance (km)
        13PR     Sidereal orbit period (sec)

        # TODO: Better specify time delta
        # TODO: Better specify the body, from a list / dict
        # TODO: Choose reference

        """
        self._check_main()
        self._select_body(body)

        self.expect([b"Observe.*\] :"])
        self.sendline("e")

        idx, _, _ = self.expect([b"Coordinate.*\] :",
                                 b"Use previous center.*\] :"])
        if idx == 1:
            self.sendline("n")
            self.expect([b"Coordinate.*\] :"])

        self.sendline("sun")

        self.expect([b"Reference.*\] :"])
        self.sendline("eclip")

        self.expect([b"Starting.*\] :"])
        self.sendline(str(start_date))
        self.expect([b"Ending.*\] :"])
        self.sendline(str(end_date))
        self.expect([b"Output.*\] :"])
        self.sendline(delta)

        self.expect([b"Accept.*\] :"])
        self.sendline("n")
        self.expect([b"Output reference.*\] :"])
        self.sendline()
        self.expect([b"Output units.*\] :"])
        self.sendline("1")
        self.expect([b"Spreadsheet.*\] :"])
        self.sendline("yes")
        self.expect([b"Label.*\] :"])
        self.sendline("no")
        self.expect([b"Type.*\] :"])
        self.sendline()

        data = self.read_until(b"$$EOE").decode('ascii')
        ephem_str = data.partition("$$SOE")[-1].partition("$$EOE")[0].strip()
        # n_lines = len(ephem_str.splitlines())
        ephem_data = np.loadtxt(StringIO(ephem_str), delimiter=",",
                                usecols=(0, 2, 4, 5, 6, 10, 11), unpack=True)
        jd, ecc, inc, omega, argp, nu, a = ephem_data

        self.expect([b".*Select.* :"])
        self.sendline("N")
        self.expect([b"\n"])

        # return (jd, a, ecc, radians(inc), radians(omega), radians(argp),
        #         radians(nu))

        return (jd, a, ecc, inc, omega, argp, nu)

    def vectors(self, body, start_date, end_date, delta):
        """Compute position and velocity vector."""
        self._check_main()
        self._select_body(body)

        self.expect([b"Observe.*\] :"])
        self.sendline("v")

        idx, _, _ = self.expect([b"Coordinate.*\] :",
                                 b"Use previous center.*\] :"])
        if idx == 1:
            self.sendline("n")
            self.expect([b"Coordinate.*\] :"])

        self.sendline("@sun")

        self.expect([b"Reference.*\] :"])
        self.sendline("eclip")

        self.expect([b"Starting.*\] :"])
        self.sendline(str(start_date))
        self.expect([b"Ending.*\] :"])
        self.sendline(str(end_date))
        self.expect([b"Output.*\] :"])
        self.sendline(delta)

        self.expect([b"Accept.*\] :"])
        self.sendline("n")
        self.expect([b"Output reference.*\] :"])
        self.sendline("J2000")
        self.expect([b"Corrections.* :"])
        self.sendline("1")
        self.expect([b"Output units.*\] :"])
        self.sendline("1")
        self.expect([b"Spreadsheet.*\] :"])
        self.sendline("yes")
        self.expect([b"Label.*\] :"])
        self.sendline("no")
        self.expect([b"Select output table.*\] :"])
        self.sendline("2")

        data = self.read_until(b"$$EOE").decode('ascii')
        ephem_str = data.partition("$$SOE")[-1].partition("$$EOE")[0].strip()
        # n_lines = len(ephem_str.splitlines())
        ephem_data = np.loadtxt(StringIO(ephem_str), delimiter=",",
                                usecols=(0,) + tuple(range(2, 8)), unpack=True)
        jd, x, y, z, vx, vy, vz = ephem_data
        r = np.column_stack((x, y, z))
        v = np.column_stack((vx, vy, vz))

        self.expect([b".*Select.* :"])
        self.sendline("N")
        self.expect([b"\n"])

        return jd, r, v

    def _select_body(self, body):
        self.sendline(str(body))
        self.expect([b"Select .*, \?, <cr>:"])
        self.sendline("e")
        self.expect([b"\n"])

    def _check_main(self):
        idx, _, _ = self.expect([b"Horizons>"])
        if idx == -1:
            raise RuntimeError("I am lost!")


def _gatherhorizonsdata(delta="1d", scale=15):

    print("Building orbit for planets from JPL Horizons...")

    start_date = datetime(2024, 5, 7)
    end_date = datetime(2025, 5, 6)

    jpl = Horizons()
    rlist = []
    rmins = []

    plnts = ("MERCURY", "VENUS", "EARTH")

    for planet in plnts:

        print("     > {}".format(planet))

        # Planet state and velocity vectors
        jd, r, v = jpl.vectors(getattr(jpl, planet), start_date, end_date,
                               delta)

        jd, a, ecc, inc, omega, argp, nu = jpl.elements(getattr(jpl, planet),
                                                        start_date,
                                                        end_date,
                                                        delta)

        print(jd, a, ecc, inc, omega, argp, nu)

        # set AU in kilometers
        AU = 1  # consts.AU / 1000  # 149597871.

        scale = 1

        rv = [math.sqrt(x ** 2 + y ** 2 + z ** 2) for (x, y, z) in r]
        xy = [((x / AU) * scale, (y / AU) * scale) for (x, y) in r[:, 0:2]]

        i = rv.index(min(rv))

        rmin = math.degrees(math.atan2(r[i, 1], r[i, 0]))

        rmins.append(rmin)
        rlist.append(xy)

    return rlist, rmins, jd.tolist()


def _gatherorbitdata(delta="1d", scale=15, verbose=False):

    # print("Building orbit for planets with SPICE...")

    spice.kclear()

    # Load the kernels that this program requires.
    this_dir = os.path.dirname(os.path.realpath(__file__))
    spice.furnsh(os.path.join(this_dir, 'epys.mk'))

    # convert starting epoch to ET
    et0 = spice.str2et('2024/05/07 00:00')
    rate = 24 * 2  # Every 30 mins
    days = [(et0 + (day * (86400 / rate))) for day in range(366 * rate)]

    # internal variables and constants
    planets = ("MERCURY", "VENUS", "EARTH")
    AU = consts.AU / 1000.  # AU [km]
    argps = []
    argpxys = []
    xyvecs = []
    nuvecs = []

    for planet in planets:

        # print("     > {}".format(planet))

        dates = []
        rvec = []   # vector of centric radii
        xyvec = []  # vector of (x,y) coordinates
        nuvec = []  # vector of nu (True Anomaly) values
        incvec = []  # vector of inclination values

        for et in days:

            if verbose:
                print('ET Seconds Past J2000: {}'.format(et))

            # Compute the apparent state of MERCURY as seen from
            # the SUN in ECLIPJ2000
            starg, ltime = spice.spkezr(planet, et, 'ECLIPJ2000',
                                        'NONE', 'SUN')

            x, y, z, vx, vy, vz = [el / AU * scale for el in starg]
            r = math.sqrt(x ** 2 + y ** 2 + z ** 2)

            if verbose:
                print('\nApparent state of MERCURY as seen from',
                      ' Sun in the J2000:')
                print(' X = {:10.4f} km (LT+S)'.format(x))
                print(' Y = {:10.4f} km (LT+S)'.format(y))
                print(' Z = {:10.4f} km (LT+S)'.format(z))
                print('VX = {:10.4f} km/s (LT+S)'.format(vx))
                print('VY = {:10.4f} km/s (LT+S)'.format(vy))
                print('VZ = {:10.4f} km/s (LT+S)'.format(vz))

            # calculate orbital elements from the starg state vector
            elts = spice.oscelt(starg, et, planetmu('Sun'))

            # define a solver for Kepler's equation
            ks = pyasl.MarkleyKESolver()

            # solve for the Eccentric Anomaly (E) with the
            # Mean Anomaly (M = elts[5]) and the
            # Eccentricity (ecc = elts[1])
            E = ks.getE(elts[5], elts[1])

            # calculate the True Anomaly (nu) from E and ecc (elts[1])
            nuarg1 = math.sqrt(1 - elts[1]) * math.cos(E / 2)
            nuarg2 = math.sqrt(1 + elts[1]) * math.sin(E / 2)

            # atan2 in python needs the arguments as (y,x)
            # rather than (x,y) ...?
            nu = 2 * math.atan2(nuarg2, nuarg1)

            rvec.append(r)  # append r for each day
            xyvec.append((x, y))  # append (x,y) coords for each day
            nuvec.append(nu)  # append True anomaly for each day

            # build date in ISO format
            date = '{} {}'.format(spice.et2utc(et, 'ISOC', 0).split('T')[0],
                                  spice.et2utc(et, 'ISOC', 0).split('T')[1])
            dates.append(date)  # append date for each day
            incvec.append(elts[2])  # append inc. for each day (rads)

            # print(date, nu * spice.dpr(), x, y, z, r, elts[0])

        # for this planet find the argument of pericenter (argp):
        # find the index of the min. r value for calculated range.
        argpi = rvec.index(min(rvec))

        # calculate argp x and y values and argp using atan2
        argpxy = (xyvec[argpi][0], xyvec[argpi][1] * math.cos(incvec[argpi]))
        argp = math.degrees(math.atan2(argpxy[1], argpxy[0]))

        argpxys.append(argpxy)  # append argp (x,y) coords.
        argps.append(argp)  # append argp
        xyvecs.append(xyvec)  # append (x,y) coords. vector
        nuvecs.append(nuvec)  # append true anomaly vector

    spice.kclear()

    return days, dates, xyvecs, argps, argpxys, nuvecs


def _frmline(a, b, c, d, line_op=1.0):
    return svg.Line(a, b, c, d, stroke_width="0.15pt",
                    stroke_dasharray="2, 2", stroke_opacity=line_op)


def _gradient(id, colors, gradrot, rotang, x, y):

    # TODO: Fix the gradient rotation...

    # print("Building gradient for {}".format(id))

    xp = x * math.cos(math.radians(rotang)) - \
        y * math.sin(math.radians(rotang))
    yp = x * math.sin(math.radians(rotang)) + \
        y * math.cos(math.radians(rotang))

    return svg.SVG("linearGradient",
                   svg.SVG("stop", stop_color=colors[0], stop_opacity=1,
                           offset="40%"),
                   svg.SVG("stop", stop_color=colors[1], stop_opacity=1,
                           offset="60%"),
                   x1="0%", y1="0%", x2="100%", y2="0%",
                   spreadMethod="pad",
                   id="{}Grad".format(id),
                   gradientTransform="rotate({}, {}, {})".format(45, xp, yp))


def _outerframe(date, frmSize=15, frm_op=0.5, diag_scl=0.65, mpoargp=False,
                frm_font_size=3, frm_ticks=8, frm_miniticks=False):

    # print("Building outer frame...")
    callerfunc = inspect.stack()[1][3]

    frmSize = frmSize * 1.2

    frm = svg.LineAxis(frmSize, 0, frmSize, 2 * math.pi, 0, 2 * math.pi)
    frm.text_start = -2.5
    frm.text_angle = 180.
    frm.text_attr["font-size"] = frm_font_size
    frm.text_attr["opacity"] = frm_op
    frm.attr["stroke_opacity"] = frm_op
    frm.ticks = [x * 2 * math.pi / frm_ticks for x in range(frm_ticks)]
    if callerfunc == 'planetsplot':
        frm.labels = lambda x: "%g" % (x * 180 / math.pi)
    else:
        frm.labels = False
    if frm_miniticks:
        frm_miniticks = [x * 2 * math.pi / frm_ticks / 9 for x in
                         range(frm_ticks * 9)]
    frm.miniticks = frm_miniticks

    # Makes a circle out of the Line Axis.
    frm_plot = svg.Fig(frm, trans="x*cos(y), x*sin(y)")

    # Draw the vertical ...
    xs = 0.9
    frmLine1 = _frmline(0, -frmSize * xs, 0, frmSize * xs, line_op=frm_op)
    # ... and horizontal frame lines through the sun.
    frmLine2 = _frmline(-frmSize * xs, 0, frmSize * xs, 0, line_op=frm_op)
    # Draw the diagonal frame lines.
    frmLine3 = _frmline(-frmSize * diag_scl, -frmSize * diag_scl,
                        frmSize * diag_scl, frmSize * diag_scl,
                        line_op=frm_op)
    frmLine4 = _frmline(-frmSize * diag_scl, frmSize * diag_scl,
                        frmSize * diag_scl, -frmSize * diag_scl,
                        line_op=frm_op)

    # And there was light...
    sun_ball = _sun()

    # Metadata
    callerfunc = inspect.stack()[1][3]
    if callerfunc == 'planetsplot':
        titletag = 'Planetary Constellation Mid-Season'
    if callerfunc == 'mpoplot':
        titletag = 'MPO Orbit Mid-Season'

    refdata = svg.Fig()
    textop = 6.0

    # TODO: fix the messy way the placement x, y are defined.
    metatitle = svg.Text(-frmSize - 3.5, frmSize + 3, titletag,
                         font_size=frm_font_size, opacity=textop,
                         text_anchor="start")
    metadate = svg.Text(-frmSize - 3.5, frmSize + 1, "{}".format(date),
                        font_size=frm_font_size, opacity=textop,
                        text_anchor="start")
    if callerfunc == 'mpoplot' and mpoargp:
        metaargp = svg.Text(-frmSize - 3.5, frmSize - 1,
                            "Arg. Periherm: {:6.1f}degsym".format(mpoargp),
                            font_size=frm_font_size, opacity=textop,
                            text_anchor="start")
    else:
        metaargp = svg.Fig()

    if callerfunc == 'planetsplot':
        xy = (-frmSize + 6.8, -frmSize - 1.8)
        reforb, grad = _planetdiag("MERCURY", xy, 0)

        reforbtext1 = svg.Text(-frmSize + 4, -frmSize - 2, "Descending",
                               font_size=frm_font_size, opacity=textop,
                               text_anchor="end")

        reforbtext2 = svg.Text(-frmSize + 9, -frmSize - 2.0, "Ascending",
                               font_size=frm_font_size, opacity=textop,
                               text_anchor="start")
        refdata = svg.Fig(reforb, reforbtext1, reforbtext2)

    return svg.Fig(frm_plot,
                   svg.Fig(frmLine1, frmLine2, frmLine3, frmLine4),
                   sun_ball,
                   metatitle,
                   metadate,
                   metaargp,
                   refdata)


def _orbitdot(a, b, theta, r_dot_adj=0.1, color="#C8C5E2", r_dot_size=0.6,
              rot_x=0.0, rot_y=0.0, dot_op=1.0, dot_str_op=1.0):
    if theta > 180:
        r_dot_adj = r_dot_adj * -1.0
    r_dot = _rellipse(a, b, theta)  # +r_dot_adj*sin(theta)
    r_trans = svg.rotate(theta, rot_x, rot_y)
    # print(r_dot)
    ret_dot = svg.Fig(svg.Dots([(r_dot, 0)],
                      svg.make_symbol("dot_{}_{}".format(theta, color),
                                      fill=color, fill_opacity=dot_op,
                                      stroke="black", stroke_width="0.15pt",
                                      stroke_opacity=dot_str_op),
                      r_dot_size, r_dot_size), trans=r_trans)
    #print(defs)
    return ret_dot


def _planetdiag(name, rpos, rotang=0.0, orb_scl=1.0, frmSizecl=10.0,
                diag_op=1.0):

    # print("Building {} diagram...".format(name))

    colors = []

    if name == "MERCURY":
        colors = ["#C8C5E2", "#373163"]
    if name == "VENUS":
        diag_op = 0.4
        colors = ["#EDE051", "#393506"]
    if name == "EARTH":
        colors = ["#00AFEF", "#003C52"]

    # colorurl="url(#{}Grad)".format(name[0:4].lower())

    # Scale the position vector ...
    # rpos = [x*frmSizecl*orb_scl for x in rpos]

    # Simplify ...
    r_x = rpos[0]
    r_y = rpos[1]

    gradrot = math.degrees(math.atan2(r_y, r_x))

    # Build a white ball for background ...
    ball_bg = _planetdot(name, rpos, r_dot_size=2.0 * orb_scl,
                         dot_color="white", dot_str_op=diag_op)
    # ... and a color ball for foreground.
    ball_fg = _planetdot(name, rpos, r_dot_size=2.0 * orb_scl,
                         dot_color=colors[0], dot_op=diag_op,
                         dot_str_op=diag_op)

    # Stack coloured ball on top of white background ball...
    ball = svg.Fig(ball_bg, ball_fg)

    grad = _gradient(name[0:4].lower(), colors, gradrot, rotang, r_x, r_y)

    if name == "MERCURY":

        # print("Buidling MPO orbit schematic...")

        # MPO line scaling factor
        mpo_line_sf = 2.0

        # MPO line start and end points
        mpo_line_st = r_x - orb_scl * mpo_line_sf
        mpo_line_en = r_x + orb_scl * mpo_line_sf * 0.720811474

        node_size = 0.15

        x1 = mpo_line_st - node_size
        x2 = mpo_line_st + node_size
        y1 = r_y - node_size
        y2 = r_y + node_size
        dec_node = svg.Fig(svg.Rect(x1=x1, y1=y1, x2=x2, y2=y2, fill="black"),
                           trans=svg.rotate(0, mpo_line_st, r_y))

        x1 = mpo_line_en - node_size
        x2 = mpo_line_en + node_size
        y1 = r_y - node_size
        y2 = r_y + node_size
        asc_node = svg.Fig(svg.Rect(x1=x1, y1=y1, x2=x2, y2=y2, fill="black"),
                           trans=svg.rotate(45, mpo_line_en, r_y))

        mpo_line = svg.Fig(svg.Line(mpo_line_st, r_y, mpo_line_en, r_y,
                           stroke_width="0.15pt",),
                           asc_node,
                           dec_node,
                           trans=svg.rotate(-rotang, r_x, r_y))

        # r_trans = rotate(theta, 0, 0)
        ball.d.append(svg.Fig(mpo_line))

    return svg.Fig(ball, trans=svg.rotate(rotang, 0, 0)), grad


def _planetdot(name, rpos, dot_color="#C8C5E2", r_dot_size=0.6,
               dot_op=1.0, dot_str_op=1.0):

    r_x = rpos[0]
    r_y = rpos[1]
    cname = dot_color.replace("#", "")

    ret_dot = svg.Fig(svg.Dots([(r_x, r_y)],
                      svg.make_symbol("dot_{}_{}".format(name, cname),
                                      fill=dot_color, fill_opacity=dot_op,
                                      stroke="black", stroke_width="0.15pt",
                                      stroke_opacity=dot_str_op),
                      r_dot_size, r_dot_size))

    return ret_dot


def _planetdotang(a, b, theta, r_dot_adj=0.23, dot_color="#C8C5E2",
                  r_dot_size=0.6, rot_x=0.0, rot_y=0.0, dot_op=1.0,
                  dot_str_op=1.0):
    if theta < 180:
        r_dot_adj = r_dot_adj * -1.0
    r_dot = _rellipse(a, b, theta)
    r_trans = svg.rotate(theta, rot_x, rot_y)
    # print(r_dot)
    ret_dot = svg.Fig(svg.Dots([(r_dot, 0)],
                      svg.make_symbol("dot_{}_{}".format(theta, dot_color),
                                      fill=dot_color, fill_opacity=dot_op,
                                      stroke="black", stroke_width="0.15pt",
                                      stroke_opacity=dot_str_op),
                      r_dot_size, r_dot_size), trans=r_trans)
    # print(theta)
    # print(r_dot*cos(radians(theta)), r_dot*sin(radians(theta)))
    return ret_dot


def _rellipse(a, b, theta):
    rret = (b ** 2) / (a - math.sqrt(a ** 2 - b ** 2) *
                       math.cos(math.radians(180 - theta)))
    return rret


def _sun(id="Sun", posx=0, posy=0, size=1.5, fill="yellow",
         stroke="orange", stroke_width="0.1pt"):
    return svg.Dots([(0, 0)], svg.make_symbol(id, stroke=stroke,
                    fill=fill, stroke_width=stroke_width), size, size)


def planetsplot(userdates=None, delta="1d", master_scale=15, demo=False,
                showplots=False):
    """
    ... explain what this does...
    """

    outdir = './sample_data/output'
    # if demo:
    #     shutil.rmtree(outdir)
    #     os.makedirs(outdir)
    # else:
    #     if not os.path.exists(outdir):
    #         os.makedirs(outdir)
    #     else:
    #         print('\n   Uh-oh! The directory {} already exists.'.format(
    #             outdir))
    #         if yesno('   Do you want to replace it?'):
    #             shutil.rmtree(outdir)
    #             os.makedirs(outdir)
    #         else:
    #             return

    orbitdata = _gatherorbitdata(delta=delta, scale=master_scale)
    ets, dates, orbits, argps, argpxys, nus = orbitdata

    if userdates is None:
        userdates = dates

    if showplots:
        plt.subplot(1, 1, 1)
        for xy in orbits:
            plt.plot([x[0] for x in xy], [y[1] for y in xy],
                     'rx', label='SPICE')
        for xy in argpxys:
            plt.plot(xy[0], xy[1], 'go')
        plt.show()

    if len(orbits[0]) == len(dates) == len(ets):

        # This rotation will put the Hermean perihelion on the X-axis.
        rotang = -argps[0]

        # Load the kernels that this program requires.
        spice.kclear()
        this_dir = os.path.dirname(os.path.realpath(__file__))
        spice.furnsh(os.path.join(this_dir, 'epys.mk'))

        output_files = []

        # A graphic will be created for each 'date' in 'userdates':
        for date in userdates:

            # get the position-index of the 'et' in the 'orbitdata' list
            # of 'ets' that is closest to the 'date' in the 'userdates'
            et = spice.str2et(date)
            dx = ets.index(getclosest(ets, et))

            # -- Outer frame -------------------------------------------------

            # Opacity of degree frame and Venus graphic
            frame_op = 0.8

            # Process calendar time strings
            date = '{} {}'.format(spice.et2utc(et, 'ISOC', 0).split('T')[0],
                                  spice.et2utc(et, 'ISOC', 0).split('T')[1])
            edate, etime = date.split()
            eyear = "{}".format(edate.split('-')[0])
            emonth = "{0:02d}".format(int(edate.split('-')[1]))
            eday = "{0:02d}".format(int(edate.split('-')[2]))
            epoch = "{}/{}/{}".format(eday, emonth, eyear)
            ep_name = "{}{}{}_{}".format(eyear, emonth, eday,
                                         etime.replace(':', ''))

            frame = _outerframe(epoch, frmSize=master_scale, frm_op=frame_op)

            # -- First Point of Aires ----------------------------------------

            # merc_loan = 48.331
            # merc_argp = 29.124
            arend = svg.make_marker("fopa_arrowend", "arrow_end",
                                    fill_opacity=0.4)

            x1, y1 = 10, 0
            x2, y2 = master_scale * 1.3, 0

            fpoa = svg.Line(x1, y1, x2, y2, stroke_width=".4pt",
                            stroke_opacity=0.4, arrow_end=arend)

            xp = (x2 * math.cos(math.radians(rotang)) -
                  y2 * math.sin(math.radians(rotang)))
            yp = (x2 * math.sin(math.radians(rotang)) +
                  y2 * math.cos(math.radians(rotang)))

            fpoa_text = svg.Text(xp + 6.5, yp - 1.0, "First Point of Aries",
                                 font_size=3, opacity=0.75)
            fpoa = svg.Fig(svg.Fig(fpoa, trans=svg.rotate(rotang, 0, 0)),
                           fpoa_text)

            # -- Some containers ---------------------------------------------

            orbs = []
            circles = []
            defs = svg.SVG("defs")

            # -- Orbit circles -----------------------------------------------

            # Build the SVG for each orbit.
            for orbit in orbits:

                if orbits.index(orbit) == 1:
                    orbit_op = 0.4
                else:
                    orbit_op = 1.0

                # Mercury's orbit will have perihelion on the X-axis
                circles.append(svg.Fig(svg.Poly(orbit, stroke_width=".25pt",
                                                stroke_opacity=orbit_op),
                                       trans=svg.rotate(rotang, 0, 0)))

            # -- Planet orbs -------------------------------------------------

            points = [orbits[0][dx], orbits[1][dx], orbits[2][dx]]

            # Build the planet orb for each planet for this chart.
            for point in points:

                # Planetary inputs ...
                if points.index(point) == 0:
                    name = "MERCURY"
                    nu = math.degrees(math.atan2(point[1], point[0])) + rotang
                    if nu < 0:
                        nu = nu + 360
                    # print(nu, nu-rotang, rotang)
                    nu = "{0:03d}".format(int(nu))
                if points.index(point) == 1:
                    name = "VENUS"
                if points.index(point) == 2:
                    name = "EARTH"

                # point_r  = [x/AU for x in point]

                orb, grad = _planetdiag(name, point, rotang)

                orbs.append(orb)
                defs.append(grad)

            # -- Build final figure ------------------------------------------

            wa = master_scale * 1.5
            svgout = svg.Fig(fpoa, frame,
                             circles[0], circles[1], circles[2],
                             orbs[0], orbs[1], orbs[2]
                             ).SVG(svg.window(-wa, wa, -wa, wa))

            svgout.prepend(defs)
            out_path = os.path.join(outdir,
                                    "merc_orbit_plot_{}_{}.svg".format(
                                        ep_name, nu))
            svgout.save(out_path)
            output_files.append(out_path)

        spice.kclear()

        return output_files

    else:
        # You'll jump to hear if the epochs for all 3 planets are not equal.
        print("There is an epoch error between the planet time values...")


def mpoplot(userdates, master_scale=15, demo=False):
    """
    ... explain what this does...
    """

    outdir = '../sample_data/output'
    # if demo:
    #     shutil.rmtree(outdir)
    #     os.makedirs(outdir)
    # else:
    #     if not os.path.exists(outdir):
    #         os.makedirs(outdir)
    #     else:
    #         print('\n   Uh-oh! The directory {} already exists.'.format(
    #             outdir))
    #         if yesno('   Do you want to replace it?'):
    #             shutil.rmtree(outdir)
    #             os.makedirs(outdir)
    #         else:
    #             return

    # Clear and load the kernels that this program requires.
    spice.kclear()
    spice.furnsh('epys.mk')

    # A graphic will be created for each 'date' in 'dates':
    for date in userdates:

        et = spice.str2et(date)
        datestr = (spice.et2utc(et, 'ISOC', 0))

        # -- Outer frame -------------------------------------------------

        dist_scl = 250.0

        elts = getorbelts(date)
        arg_peri = elts[4]

        # Opacity of degree frame and Venus graphic
        frame_op = 0.5

        # # Process JD time into calendar time strings
        # datestr = spice.et2utc(et, 'ISOC', 0)
        date = '{} {}'.format(datestr.split('T')[0],
                              datestr.split('T')[1])
        edate, etime = date.split()
        eyear = "{}".format(edate.split('-')[0])
        emonth = "{0:02d}".format(int(edate.split('-')[1]))
        eday = "{0:02d}".format(int(edate.split('-')[2]))
        epoch = "{}/{}/{}".format(eday, emonth, eyear)
        ep_name = "{}{}{}".format(eyear, emonth, eday)

        frame = _outerframe(epoch, frmSize=master_scale, frm_op=frame_op,
                            mpoargp=arg_peri)

        # -- Mercury Planet --------------------------------------------------

        # tru_ano = 90
        # look_from = 270

        # x1 = "{}%".format((100*math.sin(math.radians((tru_ano+90)/2.))))
        # x2 = "{}%".format(100-(100*sin(radians((tru_ano+90)/2.))))

        angs = range(0, 360, 1)

        plt.plot(angs, ["{}".format((100 * math.sin(math.radians(x / 2))))
                        for x in angs], 'yo-')
        plt.plot(angs, ["{}".format(100 - (100 *
                        math.sin(math.radians(x / 2)))) for x in angs], 'ro-')
        # plt.show()

        stop1 = "#C8C5E2"
        # stop2 = "#373163"

        defs = svg.SVG("defs",
                       svg.SVG("linearGradient",
                               svg.SVG("stop", stop_color=stop1,
                                       stop_opacity=1, offset="45%"),
                               svg.SVG("stop", stop_color=stop1,
                                       stop_opacity=1, offset="55%"),
                               x1="0%", y1="0%", x2="100%", y2="0%",
                               spreadMethod="pad",
                               id="mercGrad")
                       )

        # defs = svg.SVG('defs',
        #                svg.SVG('radialGradient',
        #                        svg.SVG('stop',
        #                                stop_color=stop1,
        #                                stop_opacity=1,
        #                                offset='38%'),
        #                        svg.SVG('stop',
        #                                stop_color=stop2,
        #                                stop_opacity=1,
        #                                offset='40%'),
        #                        cx='50%', cy='50%',
        #                        fx='230%', fy='50%',
        #                        r='300%',
        #                        spreadMethod='pad',
        #                        id='mercGrad')
        #                )

        merc_rad = 2439.99  # km
        merc_rad_scl = merc_rad / dist_scl
        merc_ball = svg.Ellipse(0, 0, 0, merc_rad_scl, merc_rad_scl,
                                fill="url(#mercGrad)", stroke_width="0.15pt")

        # -- MPO Orbit --

        mpo_orb_ecc = 0.163229
        mpo_orb_sma = 3394.0  # km
        mpo_orb_sma_scl = mpo_orb_sma / dist_scl
        mpo_orb_smi_scl = mpo_orb_sma_scl * math.sqrt(1 - mpo_orb_ecc ** 2)

        # Make things cleaner
        a = mpo_orb_sma_scl
        b = mpo_orb_smi_scl

        mpo_orb = svg.Ellipse(-math.sqrt(a ** 2 - b ** 2), 0, 0, a, b,
                              fill="none", stroke_width="0.25pt")
        # apof = 8
        mpo_orb_apses = svg.Line(-_rellipse(a, b, 180) - 5, 0,
                                 _rellipse(a, b, 0) + 10, 0,
                                 stroke_width="0.15pt",
                                 stroke_dasharray="2, 2")

        dot_angs = range(0, 360, 20)
        dots = [_orbitdot(a, b, x, color="black") for x in dot_angs]
        mpo_orb_dots = svg.Fig()
        for dot in dots:
            mpo_orb_dots.d.append(dot)

        mpo_orb_trans = svg.rotate(arg_peri, 0, 0)
        mpo_orb_plot = svg.Fig(mpo_orb, mpo_orb_apses, mpo_orb_dots,
                               trans=mpo_orb_trans)

        # -- Direction arrow -------------------------------------------------

        dirarend = svg.make_marker("dirarrowend", "arrow_end",
                                   fill_opacity=0.2)
        dirarend.attr["markerWidth"] = 7.5

        x1, y1 = master_scale + 1, 0.4,
        x2, y2 = master_scale + 1, 1

        dirarwstrt = svg.Line(x1, y1, x2, y2, stroke_width=".4pt",
                              stroke_opacity=0.2, arrow_end=dirarend)

        dirarw = svg.Fig(dirarwstrt, trans="x*cos(y), x*sin(y)")

        # -- Apsis view ------------------------------------------------------

        apvx, apvy = master_scale + 3, -master_scale - 3
        apsisviewball = svg.Ellipse(apvx, apvy,
                                    0, merc_rad_scl * 0.25,
                                    merc_rad_scl * 0.25,
                                    fill="url(#mercGrad)",
                                    stroke_width="0.15pt")

        apsisviewlats = svg.Fig()

        for x in range(-9, 10, 3):

            hscl = math.sin(math.radians(x * 10))
            wscl = math.cos(math.radians(x * 10))

            x1 = apvx - (merc_rad_scl * 0.25 * wscl)
            y1 = apvy + (merc_rad_scl * 0.25 * hscl)
            x2 = apvx + (merc_rad_scl * 0.25 * wscl)
            y2 = apvy + (merc_rad_scl * 0.25 * hscl)

            apsisviewlats.d.append(svg.Line(x1, y1, x2, y2,
                                   stroke_width=".2pt",
                                   stroke_opacity=0.4))

        apvarend = svg.make_marker("apvarrowend",
                                   "arrow_end",
                                   fill_opacity=0.6)
        apvarend.attr["markerWidth"] = 3.0
        apvarend.attr["markerHeight"] = 3.0

        x1, y1 = apvx, apvy - 3
        x2, y2 = apvx, apvy + 3
        apsisvieworbit = svg.Line(x1, y1, x2, y2,
                                  stroke_width=".4pt",
                                  stroke_opacity=0.6,
                                  arrow_end=apvarend)

        xd = apvx
        yd = apvy + (merc_rad_scl * 0.25 * math.sin(math.radians(arg_peri)))
        apsisviewdot = svg.Fig(svg.Dots([(xd, yd)],
                                        svg.make_symbol("apsisdot",
                                                        fill="black",
                                                        fill_opacity=0.6
                                                        ),
                                        0.6, 0.6
                                        )
                               )

        apsisview = svg.Fig(apsisviewball,
                            apsisviewlats,
                            apsisvieworbit,
                            apsisviewdot)

        # -- Build final figure ----------------------------------------------

        wa = master_scale * 1.5
        svgout = svg.Fig(frame,
                         merc_ball,
                         mpo_orb_plot,
                         dirarw,
                         apsisview
                         ).SVG(svg.window(-wa, wa, -wa, wa))

        svgout.prepend(defs)

        argp = int(arg_peri)
        svgout.save(os.path.join(outdir,
                                 "mpo_orbit_plot_{}_{}.svg".format(ep_name,
                                                                   argp)
                                 )
                    )


if __name__ == '__main__':

    # I want plots for these dates...
    dates = ("2024-May-07 00:00", "2024-May-09 14:31", "2024-May-28 09:14",
             "2024-Jun-13 15:50", "2024-Jun-29 22:26", "2024-Jul-27 15:28",
             "2024-Aug-24 08:30", "2024-Sep-09 15:06", "2024-Sep-25 21:42",
             "2024-Oct-23 14:44", "2024-Nov-20 07:46", "2024-Dec-06 14:22",
             "2024-Dec-22 20:58", "2025-Jan-19 14:00", "2025-Feb-16 07:02",
             "2025-Mar-04 13:38", "2025-Mar-20 20:14", "2025-Apr-17 13:16",
             "2025-May-03 02:19")
    # dates = ("2024-May-07 00:00", "2024-May-09 14:31")

    # ... some planetary constellation plots...
    planetsplot(userdates=dates, demo=True)

    # ... and some spacecraft orbit plots.
    # mpoplot(dates, demo=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function
from setuptools import setup
from setuptools.command.test import test as testcommand
import os
import shutil
import sys
# import epys
# from multiprocessing import util

root_dir = os.path.dirname(os.path.realpath(__file__))

with open('requirements.txt') as f:
    required = f.read().splitlines()

with open('epys/__init__.py') as f:
    for line in f.readlines():
        if line.startswith('__version__'):
            version = line.split('=')[-1].strip('\n \'')

readme = open('README.rst').read()
history = open('HISTORY.rst').read().replace('.. :changelog:', '')
authors = open('AUTHORS.rst').read()
contributing = open('CONTRIBUTING.rst').read()
LONG_DESCRIPTION = """
ePYs is a python library for the manipulation, processing and plotting of
the input and output files of ESA Experiment Planning Software (EPS).

  - Free software: BSD license
  - Documentation: http://epys.rtfd.org.

**Modules**

  - **draw** makes pretty orbit graphics
  - **events** provides a series of time/date utilities
  - **maps** does things with maps and images
  - **orbit** processes mission analysis orbit files
  - **read** reads EPS/MAPPS output into useable dataframes and/or arrays
  - **utils** more utilities ...
"""


class PyTest(testcommand):
    def finalize_options(self):
        testcommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        errcode = pytest.main(self.test_args)
        sys.exit(errcode)


class Tox(testcommand):
    def finalize_options(self):
        testcommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import tox
        errcode = tox.cmdline(self.test_args)
        sys.exit(errcode)


def cleanup():
    cleanuplist = ('build', 'dist', 'epys.egg-info')
    for file in cleanuplist:
        try:
            shutil.rmtree(os.path.join(root_dir, file))
        except OSError:
            pass

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

try:
    setup(
        name='epys',
        version=version,
        description='A python library for handling EPS output.',
        long_description=LONG_DESCRIPTION,
        author='Jonathan McAuliffe',
        author_email='watch.n.learn@gmail.com',
        url='https://github.com/johnnycakes79/epys',
        packages=['epys', 'test'],
        package_data={'test': ['data/*.out']},
        include_package_data=True,
        # cmdclass={'test': Tox},
        cmdclass={'test': PyTest},
        test_suite='test',
        # tests_require=['tox'],
        tests_require=['pytest'],
        install_requires=required,
        license="BSD",
        zip_safe=False,
        keywords='epys',
        extras_require={
            'testing': ['pytest'],
        }
    )


finally:
    cleanup()
